##########################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# Copyright (C) 2003 The Ganga Project
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
##########################################################################

import os
import os.path
import string
import sys
import time
import re

from Ganga import _gangaVersion, _gangaPythonPath
from Ganga.Utility.Config.Config import getConfig
import Ganga.Runtime

def new_version_format_to_old(version):
    """
    Convert from 'x.y.z'-style format to 'Ganga-x-y-z'

    Example:
        >>> new_version_format_to_old('6.1.11')
        'Ganga-6-1-11'
    """
    return 'Ganga-'+version.replace('.', '-')



from Ganga.Utility.files import fullpath

# This code can help debugging when files aren't closed correctly and
# managing I/O

DEBUGFILES = False
MONITOR_FILES = False

if DEBUGFILES or MONITOR_FILES:
    import __builtin__
    openfiles = {}
    oldfile = __builtin__.file

    class newfile(oldfile):

        def __init__(self, *args):
            self.x = args[0]
            if DEBUGFILES:
                from Ganga.Utility.logging import getLogger
                logger = getLogger(modulename=True)
                logger.debug("init")
                logger.debug("### OPENING %s ###" % str(self.x))
            oldfile.__init__(self, *args)
            openfiles[self.x] = self

        def close(self):
            if DEBUGFILES:
                from Ganga.Utility.logging import getLogger
                logger = getLogger(modulename=True)
                logger.debug("### CLOSING %s ###" % str(self.x))
            oldfile.close(self)
            #openfiles[ self.x ] = None
            del openfiles[self.x]

    oldopen = __builtin__.open

    def newopen(*args):
        if DEBUGFILES:
            from Ganga.Utility.logging import getLogger
            logger = getLogger()
            logger.debug("NewOpen")
        return newfile(*args)
    __builtin__.file = newfile
    __builtin__.open = newopen

    def printOpenFiles():
        from Ganga.Utility.logging import getLogger
        logger = getLogger()
        logger.debug("### %d OPEN FILES: [%s]" % (len(openfiles), ", ".join(
            str(f) for f in openfiles.keys() if openfiles[f] is not None)))

    safeFiles = ['.ganga.log', '.gangarc', 'ipythonrc', 'history', 'persist']

    def safeCloseOpenFiles():
        for f in openfiles.keys():
            if f not in safeFiles:
                openfiles[f].close()


class GangaProgram(object):

    """ High level API to create instances of Ganga programs and configure/run it """

    def __init__(self, hello_string=None, argv=sys.argv):
        """ make an instance of Ganga program
        use default hello_string if not specified
        use sys.argv as arguments if not specified"""

        self.argv = argv[:]
        # record the start time.Currently we are using this in performance measurements
        # see Ganga/test/Performance tests
        self.start_time = time.time()

        if hello_string is None:
            self.hello_string = """
*** Welcome to Ganga ***
Version: %s
Documentation and support: http://cern.ch/ganga
Type help() or help('index') for online help.

This is free software (GPL), and you are welcome to redistribute it
under certain conditions; type license() for details.

""" % _gangaVersion
        else:
            self.hello_string = hello_string

        # by default enter interactive mode
        self.interactive = True

        self.default_config_file = os.path.expanduser('~/.gangarc')

        if sys.hexversion < 0x020700F0:
            version = '{0}.{1}'.format(sys.version_info[0], sys.version_info[1])
            from Ganga.Utility.logging import getLogger
            logger = getLogger()
            logger.warning('Ganga will soon be depending on Python 2.7. '
                           'You have Python {version} installed. '
                           'See https://github.com/ganga-devs/ganga/wiki/Python-2.7'.format(version=version))

    def exit(self, *msg):
        from Ganga.Utility.logging import getLogger
        logger = getLogger()
        logger.info(self.hello_string)
        for m in msg:
            logger.error(m)
        sys.exit(1)

    # parse the options

    def parseOptions(self):
        from optparse import OptionParser

        usage = self.hello_string + \
            """\nusage: %prog [options] [script] [args] ..."""

        parser = OptionParser(usage, version=_gangaVersion)

        parser.add_option("-i", dest="force_interactive", action="store_true",
                          help='enter interactive mode after running script')

        parser.add_option("--webgui", dest="webgui",  action="store_true", default='False',
                          help='starts web GUI monitoring server')

        parser.add_option("--config", dest="config_file", action="store", metavar="FILE", default=None,
                          help='read user configuration from FILE, overrides the GANGA_CONFIG_FILE environment variable. Default: ~/.gangarc')

        parser.add_option("--config-path", dest='config_path', action="store", default=None,
                          help='site/experiment config customization path, overrides the GANGA_CONFIG_PATH environment variable. The relative paths are resolved wrt to the release directory. To use a specific file you should specify the absolute path. Default: None')

        parser.add_option("-g", "--generate-config", dest='generate_config', action="store_const", const=1,
                          help='generate a default config file, backup the existing one')

        parser.add_option("-o", "--option", dest='cmdline_options', action="append", default=[], metavar='EXPR',
                          help='set configuration options, may be repeated mutiple times,'
                               'for example: -o[Logging]Ganga.Lib=DEBUG -oGangaLHCb=INFO -o[Configuration]TextShell = IPython '
                               'The VALUE of *_PATH option is prepended. To reset it use :::VALUE')

        parser.add_option("--quiet", dest="force_loglevel", action="store_const", const='ERROR',
                          help='only ERROR messages are printed')

        parser.add_option("--very-quiet", dest="force_loglevel", action="store_const", const='CRITICAL',
                          help='only CRITICAL messages are printed')

        parser.add_option("--debug", dest="force_loglevel", action="store_const", const='DEBUG',
                          help='all messages including DEBUG are printed')

        parser.add_option("--no-mon", dest='monitoring', action="store_const", const=0,
                          help='disable the monitoring loop (useful if you run multiple Ganga sessions)')

        parser.add_option("--no-prompt", dest='prompt', action="store_const", const=0,
                          help='never prompt interactively for anything except IPython (FIXME: this is not fully implemented)')

        parser.add_option("--no-rexec", dest="rexec", action="store_const", const=0,
                          help='rely on existing environment and do not re-exec ganga process'
                               'to setup runtime plugin modules (affects LD_LIBRARY_PATH)')

        parser.add_option("--test", dest='TEST', action="store_true", default=False,
                          help='run Ganga test(s) using internal test-runner. It requires GangaTest package to be installed.'
                               'Usage example: *ganga --test Ganga/test/MyTestcase* .'
                               'Refer to [TestingFramework] section in Ganga config for more information on how to configure the test runner.')

        parser.add_option("--daemon", dest='daemon', action="store_true", default=False,
                          help='run Ganga as service.')

        parser.set_defaults(force_interactive=False, config_file=None,
                            force_loglevel=None, rexec=1, monitoring=1, prompt=1, generate_config=None)
        parser.disable_interspersed_args()

        (self.options, self.args) = parser.parse_args(args=self.argv[1:])

        def file_opens(f, message):
            try:
                return open(f)
            except IOError as x:
                self.exit(message, x)

        if self.options.config_file == '':
            self.options.config_file = None
        self.options.config_file_set_explicitly = not self.options.config_file is None

        # use GANGA_CONFIG_FILE env var if it's set
        if self.options.config_file is None:
            self.options.config_file = os.environ.get(
                'GANGA_CONFIG_FILE', None)

        if self.options.config_file:
            import Ganga.Utility.files
            self.options.config_file = Ganga.Utility.files.expandfilename(
                self.options.config_file)
            open_file = file_opens(
                self.options.config_file, 'reading configuration file')
            open_file.close()
        # we run in the batch mode if a script has been specified and other
        # options (such as -i) do not force it
        if len(self.args) > 0:
            if not self.options.force_interactive:
                self.interactive = False

# Can't check here if the file is readable, because the path isn't known
#           file_opens(self.args[0],'reading script')

    @staticmethod
    def new_version():
        versions_filename = os.path.join(getConfig('Configuration')['gangadir'], '.used_versions')
        if not os.path.exists(versions_filename):
            with open(versions_filename, 'w') as versions_file:
                versions_file.write(_gangaVersion + '\n')
            return True

        with open(versions_filename, 'r+') as versions_file:
            if versions_file.read().find(_gangaVersion) < 0:
                versions_file.write(_gangaVersion + '\n')
                return True

        return False

    @staticmethod
    def rollHistoryForward():
        versions_filename = os.path.join(getConfig('Configuration')['gangadir'], '.used_versions')
        hasLoaded_newer = False
        with open(versions_filename, 'r') as versions_file:
            this_version = versions_file.read()
            split_version_info = this_version.split('.')
            if len(split_version_info) == 3:
                major = int(split_version_info[0])
                minor = int(split_version_info[1])
                debug = int(split_version_info[2])
                if major > 6:
                    hasLoaded_newer = True
                if major == 6 and minor > 1:
                    hasLoaded_newer = True
                if major == 6 and minor == 1 and debug > 13:
                    hasLoaded_newer = True
        
        old_dir = os.path.expanduser('~/.ipython-ganga')
        if 'IPYTHONDIR' in os.environ.keys():
            old_dir = os.path.abspath(os.path.expanduser(os.environ['IPYTHONDIR']))

        single_pass_file = os.path.join(old_dir, '.have_migrated')
        from Ganga.Utility.logging import getLogger
        logger = getLogger()
        logger.debug("testing: %s" % single_pass_file)

        if not os.path.exists(single_pass_file) and os.path.exists(os.path.join(old_dir, 'history')):

            logger.info("This is your first time running Ganga >=6.1.14")
            logger.info("Now attempting to migrate your IPython history file")

            try:
                from IPython.core.interactiveshell import InteractiveShell
                from IPython.core.history import HistoryManager
                from IPython.utils.path import locate_profile, get_ipython_dir
            except ImportError as err:
                logger.error("Unable to import required IPython files, can't roll forward history")
                logger.error("If you want to roll forward your old history files you'll need to do this by hand")
                logger.error("Please try running: https://gist.github.com/minrk/6003365")

            old_text_history = os.path.join(old_dir, 'history')
            new_sqlite_history = os.path.join(old_dir, "profile_default/history.sqlite")

            if not os.path.exists(os.path.join(old_dir, "profile_default")):
                os.makedirs(os.path.join(old_dir, "profile_default"))

            IPython_history = HistoryManager(hist_file=new_sqlite_history)#, shell = InteractiveShell.instance())

            with open(old_text_history) as hist_file:
                for linenum, line in enumerate(hist_file):
                    IPython_history.store_inputs(linenum, line)

            ## Set a file to indicate that we've already made this transition
            passed_file = open(single_pass_file, "w")
            passed_file.close()

            logger.info("IPython history migrated successfully.")


    @staticmethod
    def generate_config_file(config_file):
        from Ganga.GPIDev.Lib.Config.Config import config_file_as_text
        from Ganga.Utility.logging import getLogger
        logger = getLogger()

        # Old backup routine
        if os.path.exists(config_file):
            for i in range(100):
                bn = "%s.%.2d" % (config_file, i)
                if not os.path.exists(bn):
                    try:
                        os.rename(config_file, bn)
                    except Exception, err:
                        logger.error('Failed to create config backup file %s' % bn)
                        logger.error('Old file will not be overwritten, please manually remove it and start ganga with the -g option to re-generate it')
                        logger.error('Reason: %s' % str(err))
                        return
                    logger.info('Copied current config file to %s' % bn)
                    break
            else:
                config_directory = os.path.dirname(os.path.abspath(config_file))
                config_backupdir = os.path.join(config_directory, '.gangarc_backups')
                if not os.path.exists(config_backupdir):
                    os.makedirs(config_backupdir)

                datestr = "_" + time.strftime("%d.%m.%y")
                logger.info('Copying backup config files to %s' %
                            config_backupdir)
                for i in range(100):
                    old_bn = "%s.%.2d" % (config_file, i)
                    bn = os.path.basename(old_bn)
                    new_bn = bn + datestr
                    new_bn_file = os.path.join(config_backupdir, new_bn)
                    os.rename(old_bn, new_bn_file)
                bn = "%s.%.2d" % (config_file, 0)
                os.rename(config_file, bn)
                logger.info('Copied current config file to %s' % bn)
                #raise ValueError('too many bckup files')

        logger.info('Creating ganga config file %s' % config_file)
        new_config = ''

        with open(os.path.join(os.path.dirname(Ganga.Runtime.__file__), 'HEAD_CONFIG.INI'), 'r') as config_head_file:
            new_config += config_head_file.read()
        new_config += config_file_as_text()
        new_config = new_config.replace('Ganga-SVN', new_version_format_to_old(_gangaVersion))  #Add in Ganga-x-y-z format so that it is backward compatible.
        with open(config_file, 'w') as new_config_file:
            new_config_file.write(new_config)

    @staticmethod
    def print_release_notes():
        from Ganga.Utility.logging import getLogger
        from Ganga.Utility.Config.Config import getConfig
        import itertools
        logger = getLogger('ReleaseNotes')
        if getConfig('Configuration')['ReleaseNotes'] == True:
            packages = itertools.imap(lambda x: 'ganga/python/' + x, itertools.ifilter(lambda x: x != '', ['Ganga'] + getConfig('Configuration')['RUNTIME_PATH'].split(':')))
            pathname = os.path.join(os.path.dirname(__file__), '..', '..', '..', 'release', 'ReleaseNotes-%s' % _gangaVersion)

            if not os.path.exists(pathname):
                logger.warning("couldn't find release notes for version %s" % _gangaVersion)
                return

            bounding_line = '**************************************************************************************************************\n'
            dividing_line = '--------------------------------------------------------------------------------------------------------------\n'
            with open(pathname, 'r') as f:
                try:
                    notes = [l.strip() for l in f.read().replace(bounding_line, '').split(dividing_line)]
                except Exception, err:
                    logger.error('Error while attempting to read release notes')
                    logger.debug('Reason: %s' % str(err))
                    raise

            if notes[0].find(_gangaVersion) < 0:
                logger.error("Release notes version doesn't match the stated version on line 1")
                logger.error("'%s' does not match '%s'" % (_gangaVersion, notes[0]))
                return

            log_divider = '-' * 50
            note_gen = [(p, notes[notes.index(p) + 1].splitlines()) for p in packages if p in notes]
            if note_gen:
                logger.info(log_divider)
                logger.info(log_divider)
                logger.info("Release notes for version 'Ganga-%s':" % _gangaVersion)
                logger.info(log_divider)
                logger.info('')
                logger.info(log_divider)
                for p, n in note_gen:
                    logger.info(p)
                    logger.info(log_divider)
                    # logger.info('*'*len(p))
                    logger.info('')
                    for l in n:
                        logger.info(l.strip())
                    logger.info('')
                    logger.info(log_divider)
                logger.info(log_divider)

    # this is an option method which runs an interactive wizard which helps new users to start with Ganga
    # the interactive mode is not entered if -c option was used
    def new_user_wizard(self):
        from Ganga.Utility.logging import getLogger
        from Ganga.Utility.Config.Config import load_user_config, getConfig, ConfigError

        logger = getLogger('NewUserWizard')
        specified_gangadir = os.path.expanduser(os.path.expandvars(getConfig('Configuration')['gangadir']))
        specified_config = self.options.config_file
        default_gangadir = os.path.expanduser('~/gangadir')
        default_config = self.default_config_file

        if not os.path.exists(specified_gangadir) \
                and not os.path.exists(specified_config):
            logger.info('It seems that you run Ganga for the first time')
            logger.info('Ganga will send a udp packet each time you start it in order to help the development team understand how ganga is used. You can disable this in the config file by resetting [Configuration]UsageMonitoringURL=  ')

        if not os.path.exists(specified_gangadir) \
                and not os.path.exists(default_gangadir):
            logger.info('Making default gangadir: %s' % gangadir)
            try:
                os.makedirs(gangadir)
            except OSError as err:
                logger.error("Failed to create default gangadir '%s': %s" % (gangadir, err.message))
                raise
        if self.options.generate_config:
            logger = getLogger('ConfigUpdater')
            logger.info('re-reading in old config for updating...')
            load_user_config(specified_config, {})
            self.generate_config_file(specified_config)
            sys.exit(0)
        if not os.path.exists(specified_config) \
                and not os.path.exists(default_config):
            # Sleep for 1 sec to allow for most of the bootstrap to finish so
            # the user actually sees this message last
            time.sleep(3.)
            yes = raw_input('Would you like to create default config file ~/.gangarc with standard settings ([y]/n) ?\n')
            if yes == '' or yes[0:1].upper() == 'Y':
                self.generate_config_file(default_config)
                raw_input('Press <Enter> to continue.\n')
        elif self.new_version():
            self.print_release_notes()
            self.rollHistoryForward()
            # if config explicitly set we dont want to update the versions file
            # this is so that next time ganga used with the default .gangarc
            # it will still be classed as new so the default config is updated.
            # also if config explicitly set dont try to update it for new version
            # impacts hammercloud if you do?
            if not self.options.config_file_set_explicitly:
                logger = getLogger('ConfigUpdater')
                logger.info('It appears that this is the first time you have run %s' % _gangaVersion)
                logger.info('Your ganga config file will be updated.')
                logger.info('re-reading in old config for updating...')
                load_user_config(specified_config, {})
                self.generate_config_file(specified_config)

                # config file generation overwrites user values so we need to reapply the cmd line options to these user settings
                # e.g. set -o[Configuration]gangadir=/home/mws/mygangadir and the user value gets reset to the .gangarc value
                # (not the session value but the user value has precedence)
                try:
                    opts = self.parse_cmdline_config_options(self.options.cmdline_options)
                    for section, option, val in opts:
                        config = getConfig(section).setUserValue(option, val)
                except ConfigError as x:
                    self.exit('command line option error when resetting after config generation: %s' % str(x))

    @staticmethod
    def parse_cmdline_config_options(cmdline_options):
        """ Parse a list of command line config options and return a list of triplets (section,option,value).
        In case of parsing errors, raise ConfigError exception.
        """
        mpat = re.compile(r'(\[(?P<section>\S+)\]|)(?P<option>[a-zA-z0-9._/]+)=(?P<value>.+)')
        section = None

        opts = []
        for o in cmdline_options:
            rpat = mpat.match(o)
            if rpat is None:
                from Ganga.Utility.Config import ConfigError
                raise ConfigError('syntax error: "%s"' % o)
            else:
                if rpat.group('section'):
                    section = rpat.group('section')
                if section is None:
                    from Ganga.Utility.Config import ConfigError
                    raise ConfigError('section not specified: %s' % o)
                else:
                    opts.append((section, rpat.group('option'), rpat.group('value')))
        return opts

    # configuration procedure: read the configuration files, configure and
    # bootstrap logging subsystem
    def configure(self, logLevel=None):

        import Ganga.Utility.Config
        from Ganga.Utility.Config import ConfigError

        def set_cmdline_config_options(sects=None):
            try:
                opts = self.parse_cmdline_config_options(self.options.cmdline_options)
                for section, option, val in opts:
                    should_set = True
                    if not sects is None and not section in sects:
                        should_set = False
                    if should_set:
                        config = Ganga.Utility.Config.setSessionValue(section, option, val)
            except ConfigError as x:
                self.exit('command line option error: %s' % str(x))

        # set logging options
        set_cmdline_config_options(sects=['Logging'])

        # we will be reexecutig the process so for the moment just shut up
        # (unless DEBUG was forced with --debug)
        if self.options.rexec and 'GANGA_INTERNAL_PROCREEXEC' not in os.environ and not self.options.generate_config and 'GANGA_NEVER_REEXEC' not in os.environ:
            if self.options.force_loglevel != 'DEBUG':
                self.options.force_loglevel = 'CRITICAL'
        else:  # say hello
            if logLevel:
                self.options.force_loglevel = logLevel
            if self.options.force_loglevel in (None, 'DEBUG'):
                sys.stdout.write(str(self.hello_string)+'\n')

        if self.options.config_file is None or self.options.config_file == '':
            self.options.config_file = self.default_config_file

        # initialize logging for the initial phase of the bootstrap
        # will use the default, hardcoded log level in the module until
        # pre-configuration procedure is complete
        from Ganga.Utility.logging import force_global_level, getLogger

        force_global_level(self.options.force_loglevel)

        try:
            with open(self.options.config_file) as cf:
                first_line = cf.readline()
                r = re.compile('# Ganga configuration file \(\$[N]ame: (?P<version>\S+) \$\)').match(first_line)
                this_logger = getLogger("Configure")
                if not r:
                    this_logger.error('file %s does not seem to be a Ganga config file', self.options.config_file)
                    this_logger.error('try -g option to create valid ~/.gangarc')
                else:
                    cv = r.group('version').split('-')  #Version number is in Ganga-x-y-z format
                    if len(cv) == 1:
                        cv = new_version_format_to_old(cv[0]).split('-')
                    if cv[1] != '6':
                        this_logger.error('file %s was created by a development release (%s)', self.options.config_file, r.group('version'))
                        this_logger.error('try -g option to create valid ~/.gangarc')
        except IOError as x:
            # ignore all I/O errors (e.g. file does not exist), this is just an
            # advisory check
            this_logger = getLogger("Configure")
            this_logger.debug("Config File Exception: %s" % str(x))

        if self.options.config_path is None:
            try:
                self.options.config_path = os.environ['GANGA_CONFIG_PATH']
            except KeyError, err:
                self.options.config_path = ''
            if self.options.config_path is None:
                self.options.config_path = ''

        import Ganga.Utility.files
        import Ganga.Utility.util
        import inspect
        GangaRootPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))), '../..'))

        if self.options.config_path:
            self.options.config_path = Ganga.Utility.files.expandfilename(os.path.join(GangaRootPath, self.options.config_path))

        # check if the specified config options are different from the defaults
        # and set session values appropriately
        syscfg = getConfig("System")
        if self.options.config_path != syscfg['GANGA_CONFIG_PATH']:
            syscfg.setSessionValue('GANGA_CONFIG_PATH', self.options.config_path)
        if self.options.config_file != syscfg['GANGA_CONFIG_FILE']:
            syscfg.setSessionValue('GANGA_CONFIG_FILE', self.options.config_file)

        def deny_modification(name, x):
            raise Ganga.Utility.Config.ConfigError(
                'Cannot modify [System] settings (attempted %s=%s)' % (name, x))
        syscfg.attachUserHandler(deny_modification, None)
        syscfg.attachSessionHandler(deny_modification, None)

        config = getConfig("Configuration")

        # detect default user (equal to unix user name)
        import getpass
        try:
            config.options['user'].default_value = getpass.getuser()
        except Exception as x:
            raise Ganga.Utility.Config.ConfigError('Cannot get default user name' + str(x))

        # import configuration from spyware
        from Ganga.Runtime import spyware

        monConfig = getConfig("MSGMS")

        # prevent modification during the interactive ganga session
        def deny_modification(name, x):
            raise Ganga.Utility.Config.ConfigError('Cannot modify [MSGMS] settings (attempted %s=%s)' % (name, x))
        monConfig.attachUserHandler(deny_modification, None)


        # all relative names in the path are resolved wrt the _gangaPythonPath
        # the list order is reversed so that A:B maintains the typical path precedence: A overrides B
        # because the user config file is put at the end it always may override
        # everything else
        config_files = Ganga.Utility.Config.expandConfigPath(self.options.config_path, _gangaPythonPath)
        config_files.reverse()

        # read-in config files

        # FIXME: need to construct a proper dictionary - cannot use the
        # ConfigPackage directly
        system_vars = {}
        for opt in syscfg:
            system_vars[opt] = syscfg[opt]

        def _createpath(dir):

            def _accept(fname, p=re.compile('.*\.ini$')):
                return (os.path.isfile(fname) or os.path.islink(fname)) and p.match(fname)
            files = []
            if dir and os.path.exists(dir) and os.path.isdir(dir):
                files = [os.path.join(dir, f) for f in os.listdir(dir) if
                         _accept(os.path.join(dir, f))]
            return string.join(files, os.pathsep)

        def _versionsort(s, p=re.compile(r'^(\d+)-(\d+)-*(\d*)')):
            m = p.match(s)
            if m:
                if m.group(3) == '':
                    return int(m.group(1)), int(m.group(2)), 0
                else:
                    return int(m.group(1)), int(m.group(2)), int(m.group(3))
            if s == 'SVN':
                return 'SVN'
            return None

        if "GANGA_SITE_CONFIG_AREA" in os.environ:
            this_dir = os.environ['GANGA_SITE_CONFIG_AREA']
            if os.path.exists(this_dir) and os.path.isdir(this_dir):
                dirlist = sorted(os.listdir(this_dir), key=_versionsort)
                dirlist.reverse()
                gangaver = _versionsort(new_version_format_to_old(_gangaVersion).lstrip('Ganga-')) #Site config system expects x-y-z version encoding
                for d in dirlist:
                    vsort = _versionsort(d)
                    if vsort and ((vsort <= gangaver) or (gangaver is 'SVN')):
                        select = os.path.join(this_dir, d)
                        config_files.append(_createpath(select))
                        break
        if os.path.exists(self.options.config_file):
            config_files.append(self.options.config_file)
        Ganga.Utility.Config.configure(config_files, system_vars)

        # set the system variables to the [System] module
        # syscfg.setDefaultOptions(system_vars,reset=1)

        # activate the logging subsystem
        # user defined log level takes effect NOW
        from Ganga.Utility.logging import bootstrap
        bootstrap()

        if not self.options.monitoring:
            self.options.cmdline_options.append('[PollThread]autostart=False')

        from Ganga.Utility.logging import getLogger

        logger = getLogger()

        logger.debug('default user name is %s', config['user'])
        logger.debug('user specified cmdline_options: %s', str(self.options.cmdline_options))

        # override the config options from the command line arguments
        # the format is [section]option=value OR option=value
        # in the second case last specified section from previous options is
        # used

        set_cmdline_config_options()

        if self.options.TEST:
            # FIXME: CONFIG CHECK
            # ?? config['RUNTIME_PATH'] = ''
            config.setSessionValue('RUNTIME_PATH', 'GangaTest')

        # ensure we're not interactive if daemonised
        if self.options.daemon and self.interactive:
            logger.warning("Cannot run as a service in interactive mode. Ignoring.")
            self.options.daemon = False

        # fork ourselves so we a daemon
        if self.options.daemon:
            logger.debug("Daemonising Ganga...")
            pid = os.fork()
            if pid < 0:
                sys.exit(1)

            if pid > 0:
                sys.exit(0)

            os.umask(0)
            os.setsid()
            pid = os.fork()

            if pid > 0:
                sys.exit(0)

            # change the stdout/err
            sys.stdout.flush()
            sys.stderr.flush()

            # create a server dir
            if not os.path.exists(os.path.join(config['gangadir'], "server")):
                os.makedirs(os.path.join(config['gangadir'], "server"))

            si = file("/dev/null", 'r')
            so = file(os.path.join(
                config['gangadir'], "server", "server-%s.stdout" % (os.uname()[1])), 'a')
            se = file(os.path.join(
                config['gangadir'], "server", "server-%s.stderr" % (os.uname()[1])), 'a', 0)

            os.dup2(si.fileno(), sys.stdin.fileno())
            os.dup2(so.fileno(), sys.stdout.fileno())
            os.dup2(se.fileno(), sys.stderr.fileno())

    # initialize environment: find all user-defined runtime modules and set their environments
    # if option rexec=1 then initEnvironment restarts the current ganga process (needed for LD_LIBRARY_PATH on linux)
    # set rexec=0 if you prepare your environment outside of Ganga and you do
    # not want to rexec process
    @staticmethod
    def initEnvironment(opt_rexec):

        from Ganga.Utility.logging import getLogger
        logger = getLogger()

        logger.debug("Installing Shutdown Manager")
        from Ganga.Core.InternalServices import ShutdownManager
        ShutdownManager.install()

        import Ganga.Utility.Config
        from Ganga.Utility.Runtime import RuntimePackage, allRuntimes
        from Ganga.Core import GangaException

        logger.debug("Import plugins")
        try:
            # load Ganga system plugins...
            from Ganga.Runtime import plugins
        except Exception as x:
            logger.critical('Ganga system plugins could not be loaded due to the following reason: %s', str(x))
            logger.exception(x)
            raise GangaException(x)

        # initialize runtime packages, they are registered in allRuntimes
        # dictionary automatically
        try:
            import Ganga.Utility.files
            from Ganga.Utility.Config.Config import getConfig
            config = getConfig('Configuration')

            #if config['IgnoreRuntimeWarnings']:
            #    import warnings
            #    warnings.filterwarnings(action="ignore", category=RuntimeWarning)

            
            import inspect
            from os.path import expandvars, expanduser

            GangaRootPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))), '../..'))
            def transform(x):
                return os.path.normpath(Ganga.Utility.files.expandfilename(os.path.join(GangaRootPath,x)))

            paths = map(transform, filter(None, map(lambda x: expandvars(expanduser(x)), config['RUNTIME_PATH'].split(':'))))

            for path in paths:
                r = RuntimePackage(path)
        except KeyError, err:
            logger.debug("init KeyError: %s" % str(err))

        logger.debug("Internal_ProxReexec")

        # initialize the environment only if the current ganga process has not
        # been rexeced
        if 'GANGA_INTERNAL_PROCREEXEC' not in os.environ and 'GANGA_NEVER_REEXEC' not in os.environ:
            logger.debug('initializing runtime environment')
            # update environment of the current process
            for r in allRuntimes.values():
                try:
                    _env = r.getEnvironment()
                    if isinstance(_env, dict):
                        os.environ.update(_env)
                except Exception as err:
                    logger.error("can't get environment for %s, possible problem with the return value of getEvironment()" % r.name)
                    logger.error("Reason: %s" % str(err))
                    raise

            # in some cases the reexecution of the process is needed for LD_LIBRARY_PATH to take effect
            # re-exec the process if it is allowed in the options
            if opt_rexec:
                logger.debug('re-executing the process for LD_LIBRARY_PATH changes to take effect')
                os.environ['GANGA_INTERNAL_PROCREEXEC'] = '1'
                prog = os.path.normpath(sys.argv[0])
                logger.debug('Program: %s' % str(prog))
                logger.debug('sys.argv: %s' % str(sys.argv))
                os.execv(prog, sys.argv)

        else:
            logger.debug('skipped the environment initialization -- the processed has been re-execed and setup was done already')

        # bugfix 40110
        if 'GANGA_INTERNAL_PROCREEXEC' in os.environ:
            del os.environ['GANGA_INTERNAL_PROCREEXEC']

        from Ganga.Core.GangaThread.WorkerThreads import startUpQueues
        startUpQueues()

    # bootstrap all system and user-defined runtime modules
    @staticmethod
    def bootstrap(interactive):
        import Ganga.Utility.Config
        config = Ganga.Utility.Config.getConfig('Configuration')

        from Ganga.Core import GangaException
        from Ganga.Utility.Runtime import allRuntimes
        from Ganga.Utility.logging import getLogger
        logger = getLogger()

        for n, r in zip(allRuntimes.keys(), allRuntimes.values()):
            try:
                r.bootstrap(Ganga.GPI.__dict__)
            except Exception as err:
                logger.error('problems with bootstrapping %s -- ignored', n)
                logger.error('Reason: %s' % str(err))
                raise err
            try:
                r.loadNamedTemplates(Ganga.GPI.__dict__,
                                     Ganga.Utility.Config.getConfig('Configuration')['namedTemplates_ext'],
                                     Ganga.Utility.Config.getConfig('Configuration')['namedTemplates_pickle'])
            except Exception as err:
                logger.error('problems with loading Named Templates for %s', n)
                logger.error('Reason: %s' % str(err))
        # load user-defined plugins...

        for r in allRuntimes.values():
            try:
                r.loadPlugins()
            except Exception as err:
                logger.error("problems with loading plugins for %s -- ignored" % r.name)
                logger.error('Reason: %s' % str(err))

        from Ganga.Runtime.GPIexport import exportToGPI

        from Ganga.Utility.Plugin import allPlugins
        # make all plugins visible in GPI
        for k in allPlugins.allCategories():
            for n in allPlugins.allClasses(k):
                cls = allPlugins.find(k, n)
                if not cls._declared_property('hidden'):
                    if n != cls.__name__:
                        exportToGPI(cls.__name__, cls, 'Classes')
                    exportToGPI(n, cls, 'Classes')

        # set the default value for the plugins
        from Ganga.Utility.Config import getConfig
        default_plugins_cfg = getConfig("Plugins")

        for opt in default_plugins_cfg:
            try:
                category, tag = opt.split('_')
            except ValueError, err:
                logger.warning("do not understand option %s in [Plugins]", opt)
                logger.debug('Reason: want %s' % str(err))
            else:
                if tag == 'default':
                    try:
                        allPlugins.setDefault(category, default_plugins_cfg[opt])
                    except Ganga.Utility.Plugin.PluginManagerError as x:
                        logger.warning('cannot set the default plugin "%s": %s' % (opt, x))
                else:
                    logger.warning("do not understand option %s in [Plugins]", opt)

        # set alias for default Batch plugin (it will not appear in the
        # configuration)

        batch_default_name = Ganga.Utility.Config.getConfig('Configuration').getEffectiveOption('Batch')
        try:
            batch_default = allPlugins.find('backends', batch_default_name)
        except Exception as x:
            raise Ganga.Utility.Config.ConfigError('Check configuration. Unable to set default Batch backend alias (%s)' % str(x))
        else:
            allPlugins.add(batch_default, 'backends', 'Batch')
            exportToGPI('Batch', batch_default, 'Classes')

        from Ganga.GPIDev.Base import ProtectedAttributeError, ReadOnlyObjectError, GangaAttributeError
        from Ganga.GPIDev.Lib.Job.Job import JobError

        exportToGPI('GangaAttributeError', GangaAttributeError, 'Exceptions')
        exportToGPI('ProtectedAttributeError', ProtectedAttributeError, 'Exceptions')
        exportToGPI('ReadOnlyObjectError', ReadOnlyObjectError, 'Exceptions')
        exportToGPI('JobError', JobError, 'Exceptions')

        def license():
            'Print the full license (GPL)'
            with open(os.path.join(_gangaPythonPath, '..', 'LICENSE_GPL')) as printable:
                logger.info(printable.read())

        exportToGPI('license', license, 'Functions')
        # bootstrap credentials

        from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory
        from Ganga.GPIDev.Credentials import getCredential

        # only the available credentials are exported

        # At this point we expect to have the GridProxy already created
        # by one of the Grid plugins (LCG/NG/etc) so we search for it in creds
        # cache
        credential = getCredential(name='GridProxy', create=False)
        if credential:
            exportToGPI('gridProxy', GPIProxyObjectFactory(credential), 'Objects', 'Grid proxy management object.')

        credential = getCredential('AfsToken')
        if credential:
            exportToGPI('afsToken', GPIProxyObjectFactory(credential), 'Objects', 'AFS token management object.')

        # add built-in functions

        from Ganga.GPIDev.Persistency import export, load
        exportToGPI('load', load, 'Functions')
        exportToGPI('export', export, 'Functions')

        from Ganga.Utility.logging import getLogger

        def typename(obj):
            """Return a name of Ganga object as a string, example: typename(j.application) -> 'DaVinci'"""
            from Ganga.GPIDev.Base.Proxy import isProxy, stripProxy, implRef
            if isProxy(obj):
                if hasattr(stripProxy(obj), '_name'):
                    return stripProxy(obj)._name
                else:
                    logger = getLogger()
                    logger.error("Object %s DOES NOT have the _name parameter set" % (str(obj)))
                    #import traceback
                    #traceback.print_stack()
                    return ""
            else:
                if hasattr(obj, '_name'):
                    return obj._name
                else:
                    logger = getLogger()
                    logger.error("Object %s DOES NOT have the %s or _name parameter set" % (str(obj), str(implRef)))
                    #import traceback
                    #traceback.print_stack()
                    return ""

        def categoryname(obj):
            """Return a category of Ganga object as a string, example: categoryname(j.application) -> 'applications'"""
            from Ganga.GPIDev.Base.Proxy import isProxy, stripProxy, implRef
            if isProxy(obj):
                if hasattr(stripProxy(obj), '_category'):
                    return stripProxy(obj)._category
                else:
                    logger = getLogger()
                    logger.error("Object %s DOES NOT have the _category parameter set" % (str(obj)))
                    #import traceback
                    #traceback.print_stack()
                    return ""
            else:
                if hasattr(obj, '_category'):
                    return obj._category
                else:
                    logger = getLogger()
                    logger.error("Object %s DOES NOT have the %s or _category parameter set" % (str(obj), str(implRef)))
                    #import traceback
                    #traceback.print_stack()
                    return ""

        def plugins(category=None):
            """List loaded plugins.

            If no argument is given return a dictionary of all loaded plugins.
            Keys are category name. Values are lists of plugin names in each
            category.

            If a category is specified (for example 'splitters') return a list
            of all plugin names in this category.
            """
            from Ganga.Utility.Plugin import allPlugins
            if category:
                return allPlugins.allClasses(category).keys()
            else:
                d = {}
                for c in allPlugins.allCategories():
                    d[c] = allPlugins.allCategories()[c].keys()
                return d

        # FIXME: DEPRECATED
        def list_plugins(category):
            """List all plugins in a given category, OBSOLETE: use plugins(category)"""
            raise DeprecationWarning("use plugins('%s')" % category)

        def applications():
            """return a list of all available applications, OBSOLETE: use plugins('applications')"""
            raise DeprecationWarning("use plugins('applications')")

        def backends():
            """return a list of all available backends, OBSOLETE: use plugins('backends')"""
            raise DeprecationWarning("use plugins('backends')")
        from Ganga.GPIDev.Adapters.IPostProcessor import MultiPostProcessor

        def convert_merger_to_postprocessor(j):
            from Ganga.GPIDev.Base.Proxy import stripProxy
            if len(stripProxy(j.postprocessors).process_objects):
                logger.info('job(%s) already has postprocessors' % j.fqid)
            if stripProxy(j).merger is None:
                logger.info(
                    'job(%s) does not have a merger to convert' % j.fqid)
            if not len(stripProxy(j.postprocessors).process_objects) and stripProxy(j).merger is not None:
                mp = MultiPostProcessor()
                mp.process_objects.append(stripProxy(j).merger)
                stripProxy(j).postprocessors = mp
        exportToGPI('applications', applications, 'Functions')
        exportToGPI('backends', backends, 'Functions')
        exportToGPI('list_plugins', list_plugins, 'Functions')
        # FIXME: END DEPRECATED

        exportToGPI('typename', typename, 'Functions')
        exportToGPI('categoryname', categoryname, 'Functions')
        exportToGPI('plugins', plugins, 'Functions')
        exportToGPI('convert_merger_to_postprocessor',
                    convert_merger_to_postprocessor, 'Functions')

        import Ganga.Core
        from Ganga.Runtime.Repository_runtime import startUpRegistries
        startUpRegistries()

        # export full_print
        from Ganga.GPIDev.Base.VPrinter import full_print
        exportToGPI('full_print', full_print, 'Functions')

        logger.debug("Bootstrap Core Modules")
        # bootstrap core modules
        from Ganga.GPIDev.Base.Proxy import stripProxy
        Ganga.Core.bootstrap(stripProxy(Ganga.GPI.jobs), interactive)

        import Ganga.GPIDev.Lib.Config
        exportToGPI('config', Ganga.GPIDev.Lib.Config.config,
                    'Objects', 'access to Ganga configuration')
        exportToGPI('ConfigError', Ganga.GPIDev.Lib.Config.ConfigError, 'Exceptions')

        from Ganga.Utility.feedback_report import report

        exportToGPI('report', report, 'Functions')

        # export all configuration items, new options should not be added after
        # this point
        Ganga.GPIDev.Lib.Config.bootstrap()

        logger.debug("Post-Bootstrap hooks")
        ###########
        # run post bootstrap hooks
        for r in allRuntimes.values():
            try:
                r.postBootstrapHook()
            except Exception as err:
                logger.error("problems with post bootstrap hook for %s" % r.name)
                logger.error("Reason: %s" % str(err))

    @staticmethod
    def startTestRunner(my_args):
        """
        run the testing framework
        """

        from Ganga.Utility.logging import getLogger
        logger = getLogger()

        try:
            # Important to avoid a lot of arguments over who has locked what object,
            # the tests are quite intensive and often trip on the background
            # monitoring thread(s)
            from Ganga.Core.InternalServices.Coordinator import disableMonitoringService
            disableMonitoringService()

            from GangaTest.Framework import runner
            from GangaTest.Framework import htmlizer
            from GangaTest.Framework import xmldifferencer

            tfconfig = Ganga.Utility.Config.getConfig('TestingFramework')
            rc = 1
            if tfconfig['EnableTestRunner']:
                logger.info("Starting Ganga Test Runner")

                if not my_args:
                    logger.warning("Please specify the tests to run ( i.e. ganga --test Ganga/test )")
                    return -1

                logger.info("myargs = %s" % str(" ".join(my_args)))

                rc = runner.start(test_selection=" ".join(my_args))
            else:
                logger.info("Test Runner is disabled (set EnableTestRunner=True to enable it)")

            if rc > 0 and tfconfig['EnableHTMLReporter']:
                logger.info("Generating tests HTML reports")
                rc = htmlizer.main(tfconfig)
            elif rc > 0 and tfconfig['EnableXMLDifferencer']:
                logger.info("Generating difference HTML reports")
                rc = xmldifferencer.main(my_args)
            return rc
        except ImportError as e:
            logger.error("You need GangaTest external package in order to invoke Ganga test-runner.")
            logger.error(e)
            return -1

    # run Ganga in the specified namespace, in principle the namespace should import all names from Ganga.GPI
    # if namespace is not specified then run in __main__
    def run(self, local_ns=None):

        from Ganga.Utility.logging import getLogger
        logger = getLogger("run")
        logger.debug("Entering run")

        if self.options.webgui == True:
            from Ganga.Runtime.http_server import start_server
            start_server()

        if local_ns is None:
            import __main__
            local_ns = __main__.__dict__
        # save a reference to the Ganga namespace as an instance attribute
        self.local_ns = local_ns

        # load templates for user-defined runtime modules
        from Ganga.Utility.Runtime import allRuntimes
        for r in allRuntimes.values():
            r.loadTemplates(local_ns)

        # exec ~/.ganga.py file
        fileName = fullpath('~/.ganga.py')
        if os.path.exists(fileName):
            try:
                execfile(fileName, local_ns)
            except Exception as x:
                logger.error('Failed to source %s (Error was "%s"). Check your file for syntax errors.', fileName, str(x))
        # exec StartupGPI code
        from Ganga.Utility.Config import getConfig
        config = getConfig('Configuration')
        if config['StartupGPI']:
            # ConfigParser trims the lines and escape the space chars
            # so we have only one possibility to insert python code :
            # using explicitly '\n' and '\t' chars
            code = config['StartupGPI'].replace(
                '\\t', '\t').replace('\\n', '\n')
            exec code in local_ns

        logger.debug("loaded .ganga.py")

        # monitor the  ganga usage
        from Ganga.Runtime import spyware

        # this logic is a bit convoluted
        runs_script = len(self.args) > 0
        session_type = config['TextShell']
        if runs_script:
            if not self.interactive:
                session_type = 'batch'
                from Ganga.Utility.Config import setConfigOption
                setConfigOption('PollThread', 'forced_shutdown_policy', 'batch')
                from Ganga.Core import change_atexitPolicy
                change_atexitPolicy(False, 'batch')
            else:
                session_type += 'startup_script'

        spyware.ganga_started(session_type=session_type, interactive=self.interactive,
                              webgui=self.options.webgui,
                              script_file=runs_script, text_shell=config['TextShell'],
                              test_framework=self.options.TEST)

        if self.options.TEST:
            sys.argv = self.args
            try:
                rc = self.startTestRunner(self.args)
            except (KeyboardInterrupt, SystemExit):
                logger.warning('Test Runner interrupted!')
                import Ganga.Core.InternalServices.Coordinator
                if not Ganga.Core.InternalServices.Coordinator.servicesEnabled:
                    from Ganga.GPI import reactivate
                    reactivate()
                sys.exit(1)
            sys.exit(rc)

        if len(self.args) > 0:
            # run the script and make it believe it that it is running directly
            # as an executable (sys.argv)
            saved_argv = sys.argv
            sys.argv = self.args

            import Ganga.Utility.Runtime
            path = Ganga.Utility.Runtime.getSearchPath()
            script = Ganga.Utility.Runtime.getScriptPath(self.args[0], path)

            #if script:
            #    script_file = open(script, 'r')
            #    script_content = script_file.read()
            #    script_file.close()
            #
            #    compiled_script = compile(script_content, '<string>', 'exec')
            #
            #exec compiled_script

            if script:
                execfile(script, local_ns)
            else:
                logger.error("'%s' not found" % self.args[0])
                logger.info("Searched in path %s" % path)
                sys.exit(1)
            sys.argv = saved_argv

        # and exit unless -i was specified
        if not self.interactive:
            return
            #    sys.exit(0)

        # interactive python shell

        shell = config['TextShell']

        if shell == 'IPython':

            try:
                import IPython
            except ImportError as err:
                logger.error("Cannot load IPython class!!!")
                logger.error("Please check your environment and re-load Ganga from a working shell!")
                logger.error("Exiting Ganga now, goodbye!")
                return

            import Ganga.Utility.Config
            ipconfig = Ganga.Utility.Config.getConfig('TextShell_IPython')
            args = eval(ipconfig['args'])

            # buffering of log messages from all threads called "GANGA_Update_Thread"
            # the logs are displayed at the next IPython prompt

            from Ganga.Utility.logging import enableCaching
            enableCaching()

            ipver = IPython.__version__

            ipver_major = int(ipver[0])
            ipver_minor = int(ipver[2])

            #if ipver in ["1.2.1", "3.1.0", "3.2.0", "3.2.1", '4.0.0']:
            if ipver_major > 1 or (ipver_major == 1 and ipver_minor >= 2):
                self.check_IPythonDir()
                self.launch_IPython(local_ns, args, self._ganga_error_handler, self.ganga_prompt)
            else:
                print("Unknown IPython version: %s" % str(ipver))
                return

        else:
            override_credits()
            import code
            sys.displayhook = _display
            c = code.InteractiveConsole(locals=local_ns)
            c.interact()

        return

    @staticmethod
    def check_IPythonDir():
        """
        Check that the .ipython folder exists as expected as defined by IPYTHONDIR
        """

        from Ganga.Utility.logging import getLogger
        logger = getLogger()

        not_exist = False

        if 'IPYTHONDIR' in os.environ.keys():
            ipyth_dir = os.environ['IPYTHONDIR']
            os.environ['IPYTHONDIR'] = os.path.abspath(os.path.expanduser(os.environ['IPYTHONDIR']))
            #logger.warning('Environment variable IPYTHONDIR=%s exists and overrides the default history file for Ganga IPython commands', ipyth_dir)
            if not os.path.isdir(ipyth_dir):
                if os.path.exists(ipyth_dir):
                    ipyth_dir_bak = str(ipyth_dir).append('_bak')
                    logger.warning('IPYTHONDIR points to: %s which is not a directory, moving it to: %s' % (ipyth_dir, ipyth_dir_bak))
                    os.rename(ipyth_dir, ipyth_dir_bak)
                    not_exist = True
        else:
            newpath = os.path.expanduser('~/.ipython-ganga')
            oldpath = os.path.expanduser('~/.ipython')
            os.environ['IPYTHONDIR'] = newpath
            if not os.path.exists(newpath):
                if os.path.exists(oldpath):
                    logger.warning('Default location of IPython history files has changed.')
                    logger.warning('Ganga will now try to copy your old settings from %s to the new path %s.\
                            If you do not want that, quit Ganga and wipe off the content of new path: rm -rf %s/*', oldpath, newpath, newpath)
                    import shutil
                    shutil.copytree(oldpath, newpath)
                else:
                    not_exist = True

        if not_exist:
            os.makedirs(os.environ['IPYTHONDIR'])

        rc_file = os.path.join(os.environ['IPYTHONDIR'], 'ipythonrc')
	logger.debug("Checking: %s" % str(rc_file))
        if not os.path.isfile(rc_file):
            lock = open(rc_file, "w")
            lock.close()

        return None

    @staticmethod
    def _ganga_error_handler(exception_obj, etype, value, tb, tb_offset=None):
        """
        Error handler for IPython 3.x+ to identify expected Ganga exceptions or unexpected uncaught exceptions from somewhere
        """
        ## see https://ipython.org/ipython-doc/dev/api/generated/IPython.core.interactiveshell.html#IPython.core.interactiveshell.InteractiveShell.set_custom_exc
        from Ganga.Utility.logging import getLogger
        logger = getLogger(modulename=True)
        logger.error("Error: %s" % str(value))

        from Ganga.Core.exceptions import GangaException
        if not issubclass(etype, GangaException):
            logger.error("Unknown/Unexpected ERROR!!")
            #logger.error("If you're able to reproduce this please report this to the Ganga developers!")
            #logger.error("value: %s" % value)
            exception_obj.showtraceback((etype, value, tb), tb_offset=tb_offset)
        return None

    @staticmethod
    def launch_IPython(local_ns, args, error_handler, ganga_prompt):
        """
        Launch an embedded IPython session within the Ganga.GPI namespace
        """

        import IPython

        # Based on examples/Embedding/embed_class_long.py from the IPython source tree

        # First we set up the prompt
        from IPython.config.loader import Config
        cfg = Config()
        cfg.TerminalInteractiveShell.colors = 'LightBG'
        cfg.TerminalInteractiveShell.autocall = 0
        cfg.PlainTextFormatter.pprint = True
        banner = exit_msg = ''
        prompt_config = cfg.PromptManager
        prompt_config.in_template = '[{time}]\nGanga In [\\#]: '
        prompt_config.in2_template = '         .\\D.: '
        prompt_config.out_template = 'Ganga Out [\\#]: '

        # Import the embed function
        from IPython.terminal.embed import InteractiveShellEmbed

        ## Check which version of IPython we're running
        ipver = IPython.__version__
        ipver_major = int(ipver[0])
        if ipver_major > 2:
            ipshell = InteractiveShellEmbed(argv=args, config=cfg, banner1=banner, exit_msg=exit_msg)
        else:
            ipshell = InteractiveShellEmbed(config=cfg, banner1=banner, exit_msg=exit_msg)

        ipshell.set_custom_exc((Exception,), error_handler)

        # buffering of log messages from all threads called "GANGA_Update_Thread"
        # the logs are displayed at the next IPython prompt
        from Ganga.Utility.logging import enableCaching
        enableCaching()

        ipshell.set_hook("pre_prompt_hook", ganga_prompt)

        from Ganga.Runtime.IPythonMagic import magic_ganga
        ipshell.define_magic('ganga', magic_ganga)

        from Ganga.Utility.Config.Config import getConfig
        config = getConfig('Configuration')

        ipshell.confirm_exit = config['confirm_exit']

        # Launch embedded shell
        ipshell(local_ns=local_ns, global_ns=local_ns)

    @staticmethod
    def ganga_prompt(dummy=None):
        from Ganga.Utility.logging import cached_screen_handler
        if cached_screen_handler:
            cached_screen_handler.flush()

        credentialsWarningPrompt = ''
        # alter the prompt only when the internal services are disabled
        from Ganga.Core.InternalServices import Coordinator
        if not Coordinator.servicesEnabled:
            invalidCreds = Coordinator.getMissingCredentials()
            if invalidCreds:
                credentialsWarningPrompt = '[%s required]' % ','.join(invalidCreds)
            if credentialsWarningPrompt:  # append newline
                 credentialsWarningPrompt += '\n'

        return credentialsWarningPrompt



#
# $Log: not supported by cvs2svn $
# Revision 1.11.4.1  2009/07/08 11:18:21  ebke
# Initial commit of all - mostly small - modifications due to the new GangaRepository.
# No interface visible to the user is changed
#
# Revision 1.11  2009/04/28 13:37:12  kubam
# simplified handling of logging filters
#
# Revision 1.15  2009/07/20 14:13:44  moscicki
# workaround for wierd OSX execv behaviour (from Ole Weidner)
#
# Revision 1.14  2009/06/10 14:53:05  moscicki
# fixed bug #51592: Add self to logger
#
# Revision 1.13  2009/06/09 10:44:55  moscicki
# removed obsolete variable
#
# Revision 1.12  2009/06/08 15:48:17  moscicki
# fix Ganga to work with newer versions of ipython (-noautocall option was removed in newer ipython versions)
#
# Revision 1.11  2009/04/28 13:37:12  kubam
# simplified handling of logging filters
#
# Revision 1.10  2009/02/02 13:43:26  moscicki
# fixed: bug #44934: Didn't create .gangarc on first usage
#
# Revision 1.9  2008/11/27 15:49:03  moscicki
# extra exception output if cannot load the plugins...
#
# Revision 1.8  2008/11/21 16:34:22  moscicki
# bug #43917: Implement Batch backend as alias to default backend at a given site
#
# Revision 1.7  2008/10/23 15:24:04  moscicki
# install the shutdown manager for atexit handlers before loading system plugins (e.g. LCG download thread registers the atexit handler using a tuple (priority,handler))
#
# Revision 1.6  2008/09/05 15:55:51  moscicki
# XML differenciater added (from Ulrik)
#
# Revision 1.5  2008/08/18 13:18:59  moscicki
# added force_status() method to replace job.fail(), force_job_failed() and
# force_job_completed()
#
# Revision 1.4  2008/08/18 10:02:15  moscicki
#
# bugfix 40110
#
# Revision 1.3  2008/08/01 15:25:30  moscicki
# typo fix
#
# Revision 1.2  2008/07/31 17:25:02  moscicki
# config templates are now in a separate directory at top level ("templates")
#
# *converted all tabs to spaces*
#
# Revision 1.1  2008/07/17 16:41:00  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.71.4.23  2008/07/03 16:11:31  moscicki
# bug #38000: Add check for old .gangarc file
#
# Revision 1.71.4.22  2008/04/03 12:55:45  kuba
# importing core plugins before initEnvironment(), this fixes
# bug #35146: GangaAtlas is not starting due to gridshell call in __init__.py
#
# Revision 1.71.4.21  2008/04/01 14:08:03  roma
# automatic config file template generation (Vladimir)
#
# Revision 1.71.4.20  2008/03/31 15:32:46  kubam
# use more flexible logic for hidden classes
#
# Revision 1.71.4.19  2008/03/12 17:33:32  moscicki
# workaround for broken logging system: GangaProgram.log writes directly to stderr
#
# Revision 1.71.4.18  2008/03/11 15:24:51  moscicki
# merge from Ganga-5-0-restructure-config-branch
#
# Revision 1.71.4.17.2.1  2008/03/07 13:34:38  moscicki
# workspace component
#
# Revision 1.71.4.17  2008/03/06 14:14:55  moscicki
# streamlined session options logic
# moved sanityCheck to config.bootstrap()
#
# Revision 1.71.4.16  2008/03/06 11:26:52  amuraru
# fixed .ganga.py sourcing file
#
# Revision 1.71.4.15  2008/03/05 14:53:33  amuraru
# execute ~/.ganga.py file before executing StartupGPI code
#
# Revision 1.71.4.14  2008/02/28 10:08:32  amuraru
# *** empty log message ***
#
# Revision 1.71.4.13  2008/02/21 12:11:02  amuraru
# added [Shell] section to configure internal Shell utility
#
# Revision 1.71.4.12  2008/02/06 17:04:11  moscicki
# initialize external monitoring services subsystem
#
# Revision 1.71.4.11  2007/12/18 16:51:28  moscicki
# merged from XML repository branch
#
# Revision 1.71.4.10  2007/12/18 13:05:19  amuraru
# removed coverage code from boostrap (moved in GangaTest/Framework/driver.py)
#
# Revision 1.71.4.9  2007/12/13 16:33:02  moscicki
# export more GPI exceptions
#
# Revision 1.71.4.8  2007/12/10 18:55:55  amuraru
# merged changes from Ganga 4.4.4
#
# Revision 1.71.4.7  2007/11/14 11:41:46  amuraru
# 5.0 configuration updated
#
# Revision 1.71.4.6.2.1  2007/11/13 16:26:05  moscicki
# removed obsolete migration GPI commands
#
# Revision 1.71.4.6  2007/11/08 13:21:00  amuraru
# moved testconfig option defintion to GangaTest
#
# Revision 1.71.4.5  2007/11/07 15:10:04  moscicki
# merged in pretty print and GangaList support from ganga-5-dev-branch-4-4-1-will-print branch
#
#
# Revision 1.71.4.4  2007/11/02 15:20:32  moscicki
# moved addOption() before config bootstrap
#
# Revision 1.71.4.3  2007/10/31 13:39:45  amuraru
# update to the new config system
#
# Revision 1.71.4.2  2007/10/25 11:43:11  roma
# Config update
#
# Revision 1.71.4.1  2007/10/12 13:56:26  moscicki
# merged with the new configuration subsystem
#
# Revision 1.71.6.2  2007/10/09 07:31:56  roma
# Migration to new Config
#
# Revision 1.71.6.1  2007/09/25 09:45:12  moscicki
# merged from old config branch
#
# Revision 1.71.8.1  2007/10/30 12:12:08  wreece
# First version of the new print_summary functionality. Lots of changes, but some known limitations. Will address in next version.
#
# Revision 1.76  2007/11/26 12:13:27  amuraru
# decode tab and newline characters in StartupGPI option
#
# Revision 1.75  2007/11/05 12:33:54  amuraru
# fix bug #30891
#
# Revision 1.74  2007/10/29 14:04:08  amuraru
#  - added free disk space checking in [PollThread] configuration template
#  - added an extra check not to attempt the shutdown of the repository if this has already been stopped
#  - save the Ganga namespace as an attribute in Ganga.Runtime._prog
#
# Revision 1.73  2007/10/10 14:47:46  moscicki
# updated doc-strings
#
# Revision 1.72  2007/09/25 15:12:04  amuraru
#
# usa GANGA_CONFIG_FILE  environment variable to set the user config file
#
# Revision 1.71  2007/09/11 16:54:52  amuraru
# catch the TestRunner KeyboardInterrupt
#
# Revision 1.70  2007/09/11 14:28:29  amuraru
# implemented FR #28406 to allow definition of GPI statements to be executed at
# startup
#
# Revision 1.69  2007/08/27 10:47:30  moscicki
# overriden credits() and copyright() (request #21906)
#
# Revision 1.68  2007/08/22 15:58:55  amuraru
# Runtime/bootstrap.py
#
# Revision 1.67  2007/08/14 14:47:01  amuraru
# automatically add GangaTest RT package when --test is used
#
# Revision 1.66  2007/08/13 17:22:27  amuraru
# - testing framework small fix
#
# Revision 1.65  2007/08/13 13:19:48  amuraru
# -added EnableTestRunner and EnableHTMLReported to control the testing framework in a more flexible way
# -added GANGA_CONFIG_FILE in [System] config
#
# Revision 1.64  2007/08/13 12:50:18  amuraru
# added EnableTestRunner and EnableHTMLReported to control the testing framework in a more flexible way
#
# Revision 1.63  2007/07/30 12:57:51  moscicki
# removing IPython autocall option (obsoletion of jobs[] syntax and putting jobs() as a replacement)
#
# Revision 1.62  2007/07/27 14:31:55  moscicki
# credential and clean shutdown updates from Adrian (from Ganga-4-4-0-dev-branch)
#
# Revision 1.61  2007/07/10 13:08:32  moscicki
# docstring updates (ganga devdays)
#
# Revision 1.60  2007/06/07 10:25:02  amuraru
# bug-fix: guard against environment update for RuntimePackages exposing null environment dictionary
#
# Revision 1.59  2007/06/04 14:31:22  amuraru
# record start-time of ganga session
#
# Revision 1.58  2007/06/01 08:49:08  amuraru
# Disable the autocompletion of private attributes and methods starting with _ or __
#
# Revision 1.57  2007/05/21 16:07:57  amuraru
# integrated TestingFramework into ganga itsefl (ganga --test). [TestingFramework] section in Ganga config is used to control it.
# changed Ganga.Runtime.bootstrap default log level to INFO
#
# Revision 1.56  2007/05/11 13:21:24  moscicki
# temporary functions to help getting jobs out of completing and submitting states
# force_job_completed(j): may be applied to completing jobs
# force_job_failed(j): may be applied to submitting or completing jobs
#
# Revision 1.55  2007/05/08 10:32:42  moscicki
# added short GPL license summary at startup and license() command in GPI for full print
#
# Revision 1.54.6.1  2007/06/18 07:44:57  moscicki
# config prototype
#
# Revision 1.54  2007/02/28 18:24:53  moscicki
# moved GangaException to Ganga.Core
#
# Revision 1.53  2007/02/22 13:43:19  moscicki
# pass interactive flag to Core.bootstrap
#
# Revision 1.52  2007/01/25 15:52:39  moscicki
# mergefrom_Ganga-4-2-2-bugfix-branch_25Jan07 (GangaBase-4-14)
#
# Revision 1.51.2.3  2006/12/15 17:12:37  kuba
# added spyware at startup
#
# Revision 1.51.2.2  2006/11/24 14:52:56  amuraru
# only available credentials (afs/gridproxy) are exported
#
# Revision 1.51.2.1  2006/11/24 14:22:05  amuraru
# added support for peek() function
#
# Revision 1.51  2006/10/23 10:59:41  moscicki
# initialize [Configuration]LOAD_PATH
#
# Revision 1.50  2006/10/16 12:53:13  moscicki
# fix the SCRIPTS_PATH mechanism: the . is always in the path and the session level updates are prepending to the default value... fix for bug #20332 overview: Ganga/scripts not included in SCRIPTS_PATH in Atlas.ini
#
# Revision 1.49  2006/10/04 18:16:48  moscicki
# fixed bug #20333 overview: hostname function of Ganga/Utility/util.py sometimes fails
#
# Revision 1.48  2006/09/27 16:38:31  moscicki
# changed AfsToken -> afsToken, GridProxy -> gridProxy and made them real GPI proxy objects
#
# Revision 1.47  2006/09/15 14:23:31  moscicki
# Greeting message goes to stderr (requested by UNOSAT to use Ganga in CGI scripts).
#
# Revision 1.46  2006/08/29 15:11:10  moscicki
# fixed #18084 Additonal global objects for splitters, mergers etc
#
# Revision 1.45  2006/08/29 12:51:57  moscicki
# exported GridProxy and AfsToken singleton objects to GPI
#
# Revision 1.44  2006/08/11 13:13:06  adim
# Added: GangaException as a markup base class for all exception that need to be printed in a usable way in IPython shell
#
# Revision 1.43  2006/08/09 09:07:34  moscicki
# added magic functions ('ganga')
#
# Revision 1.42  2006/07/31 12:13:43  moscicki
# depend on monitoring thread names "GANGA_Update_Thread" to do message buffering in IPython
#
# Revision 1.41  2006/07/27 20:21:24  moscicki
# - fixed option parsing
# - pretty formatting of known exceptions in IPython (A.Muraru)
#
# Revision 1.40  2006/06/21 11:43:00  moscicki
# minor fix
#
# Revision 1.39  2006/03/14 14:53:14  moscicki
# updated comments
#
# Revision 1.38  2006/03/09 08:41:52  moscicki
# --gui option and GUI integration
#
# Revision 1.37  2006/02/13 15:21:25  moscicki
# support for cached logging messages at interactive prompt (messages from monitoring thread are cached in IPython environment and printed at the next prompt)
#
# Revision 1.36  2006/02/10 14:16:00  moscicki
# fixed bugs:
# #13912        Cannot use tilde to give location of INI file
# #14436 problem with -o option at the command line and setting config default for properties
#
# exported ConfigError to GPI
# docstring updates
#
# Revision 1.35  2005/11/25 09:57:37  moscicki
# exported TreeError exception
#
# Revision 1.34  2005/11/14 14:47:38  moscicki
# jobtree added
#
# Revision 1.33  2005/11/14 10:29:10  moscicki
# support for default plugins
# temporary hack for GUI-specific monitoring
#
# Revision 1.32  2005/11/01 11:21:37  moscicki
# support for export/load (KH)
#
# Revision 1.31  2005/10/14 12:54:38  moscicki
# ignore I/O exceptions while checking for ganga3 config file
#
# Revision 1.30  2005/10/12 13:35:23  moscicki
# renamed _gangadir into gangadir
#
# Revision 1.29  2005/10/07 15:08:45  moscicki
# renamed __Ganga4__ into _gangadirm .ganga4 into .gangarc
# added sanity checks to detect old (Ganga3) config files
# added config-path mechanism
#
# Revision 1.28  2005/10/07 08:27:00  moscicki
# all configuration items have default values
#
# Revision 1.27  2005/09/22 12:48:14  moscicki
# import fix
#
# Revision 1.26  2005/09/21 09:12:50  moscicki
# added interactive displayhooks based on obj._display (if exists)
#
# Revision 1.25  2005/08/26 10:12:06  moscicki
# added [System] section (write-protected) with GANGA_VERSION and GANGA_PYTHONPATH (new)
#
# Revision 1.24  2005/08/24 15:24:11  moscicki
# added docstrings for GPI objects and an interactive ganga help system based on pydoc
#
#
#



import os
## NB parseCommandLine first then import Dirac!!
from DIRAC.Core.Base.Script import parseCommandLine
parseCommandLine()
from DIRAC.Interfaces.API.Dirac import Dirac
dirac=Dirac()
dirac.getFile('###LFN###', os.getcwd())



import re
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.Utility.Config import getConfig


class LCGRequirements(GangaObject):

    '''Helper class to group LCG requirements.

    See also: JDL Attributes Specification at http://cern.ch/glite/documentation
    '''

    _schema = Schema(Version(1, 2), {
        'software': SimpleItem(defvalue=[], typelist=['str'], sequence=1, doc='Software Installations'),
        'nodenumber': SimpleItem(defvalue=1, doc='Number of Nodes for MPICH jobs'),
        'memory': SimpleItem(defvalue=0, doc='Mininum available memory (MB)'),
        'cputime': SimpleItem(defvalue=0, doc='Minimum available CPU time (min)'),
        'walltime': SimpleItem(defvalue=0, doc='Mimimum available total time (min)'),
        'ipconnectivity': SimpleItem(defvalue=False, doc='External connectivity'),
        'allowedCEs': SimpleItem(defvalue='', doc='allowed CEs in regular expression'),
        'excludedCEs': SimpleItem(defvalue='', doc='excluded CEs in regular expression'),
        'datarequirements': SimpleItem(defvalue=[], typelist=['dict'], sequence=1, doc='The DataRequirements entry for the JDL. A list of dictionaries, each with "InputData", "DataCatalogType" and optionally "DataCatalog" entries'),
        'dataaccessprotocol': SimpleItem(defvalue=['gsiftp'], typelist=['str'], sequence=1, doc='A list of strings giving the available DataAccessProtocol protocols'),
        'other': SimpleItem(defvalue=[], typelist=['str'], sequence=1, doc='Other Requirements')
    })

    _category = 'LCGRequirements'
    _name = 'LCGRequirements'

    def __init__(self):

        super(LCGRequirements, self).__init__()

    def merge(self, other):
        '''Merge requirements objects'''

        if not other:
            return self

        merged = LCGRequirements()
        for name in ['software', 'nodenumber', 'memory', 'cputime', 'walltime', 'ipconnectivity', 'allowedCEs', 'excludedCEs', 'datarequirements', 'dataaccessprotocol', 'other']:

            attr = ''

            try:
                attr = getattr(other, name)
            except KeyError as e:
                pass

            if not attr:
                attr = getattr(self, name)
            setattr(merged, name, attr)

        return merged

    def convert(self):
        '''Convert the condition in a JDL specification'''

        import re

        requirements = [
            'Member("%s",other.GlueHostApplicationSoftwareRunTimeEnvironment)' % sw for sw in self.software]
        if self.memory:
            requirements += ['other.GlueHostMainMemoryVirtualSize >= %s' %
                             str(self.memory)]
        if self.cputime:
            requirements += [
                '(other.GlueCEPolicyMaxCPUTime >= %s || other.GlueCEPolicyMaxCPUTime == 0)' % str(self.cputime)]
        if self.walltime:
            requirements += [
                '(other.GlueCEPolicyMaxWallClockTime >= %s || other.GlueCEPolicyMaxWallClockTime == 0)' % str(self.walltime)]
        if self.ipconnectivity:
            requirements += ['other.GlueHostNetworkAdapterOutboundIP==true']
        requirements += self.other

        config = getConfig('LCG')

        # retrieve allowed_ces and excluded_ces from LCGRequirement object and
        # the config['Allowed/ExcludedCEs']
        allowed_ces = []
        excluded_ces = []

        # from Ganga configuration
        if config['AllowedCEs']:
            ce_req = config['AllowedCEs'].strip()
            allowed_ces += re.split('\s+', ce_req)

        if config['ExcludedCEs']:
            ce_req = config['ExcludedCEs'].strip()
            excluded_ces += re.split('\s+', ce_req)

        # from LCGRequirements object
        # if string starts with '+', it means the requirement to be appeneded
        re_append = re.compile('^(\++)\s*(.*)')
        try:
            ce_req = self.allowedCEs.strip()
            if ce_req:
                m = re_append.match(ce_req)
                if m:
                    allowed_ces += re.split('\s+', m.group(2))
                else:
                    allowed_ces = re.split('\s+', ce_req)
        except KeyError as e:
            pass

        try:
            ce_req = self.excludedCEs.strip()
            if ce_req:
                m = re_append.match(ce_req)
                if m:
                    excluded_ces += re.split('\s+', m.group(2))
                else:
                    excluded_ces = re.split('\s+', ce_req)
        except KeyError as e:
            pass

        # composing the requirements given the list of allowed_ces and
        # excluded_ces
        if allowed_ces:
            requirements += ['( %s )' % ' || '.join(
                ['RegExp("%s",other.GlueCEUniqueID)' % ce for ce in allowed_ces])]

        if excluded_ces:
            #requirements += [ '(!RegExp("%s",other.GlueCEUniqueID))' % ce for ce in excluded_ces ]
            requirements += ['( %s )' % ' && '.join(
                ['(!RegExp("%s",other.GlueCEUniqueID))' % ce for ce in excluded_ces])]

        return requirements



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: CondorRequirements.py,v 1.2 2008-07-29 10:30:39 karl Exp $
###############################################################################
# File: CondorRequirements.py
# Author: K. Harrison
# Created: 051229
#
# KH - 060728 : Correction to way multiple requirements are combined
#               in convert method
#
# KH - 060829 : Typo corrected
#
# KH - 061026 : Correction for missing import (types module)
#               Correction to allow configuration values for "machine"
#               and "excluded_machine" to be either string or list
#               Correction to handling of requirement for allowed machines
#
# KH - 080729 : Updates for typing system of Ganga 5
#               Improvements to documentation for CondorRequirements properties

"""Module containing class for handling Condor requirements"""

__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "29 July 2008"
__version__ = "1.4"

from Ganga.GPIDev.Base.Objects import GangaObject
from Ganga.GPIDev.Schema import Schema, SimpleItem, Version
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger

logger = getLogger()


class CondorRequirements(GangaObject):

    '''Helper class to group Condor requirements.

    See also: http://www.cs.wisc.edu/condor/manual
    '''

    _schema = Schema(Version(1, 0), {
        "machine": SimpleItem(defvalue="", typelist=["str", "list"],
                              doc="""
Requested execution hosts, given as a string of space-separated names:
'machine1 machine2 machine3'; or as a list of names:
[ 'machine1', 'machine2', 'machine3' ]
""" ),
        "excluded_machine": SimpleItem(defvalue="",
                                       typelist=["str", "list"],
                                       doc="""
Excluded execution hosts, given as a string of space-separated names:
'machine1 machine2 machine3'; or as a list of names:
[ 'machine1', 'machine2', 'machine3' ]
""" ),
        "opsys": SimpleItem(defvalue="LINUX", doc="Operating system"),
        "arch": SimpleItem(defvalue="INTEL", doc="System architecture"),
        "memory": SimpleItem(defvalue=400, doc="Mininum physical memory"),
        "virtual_memory": SimpleItem(defvalue=400,
                                     doc="Minimum virtual memory"),
        "other": SimpleItem(defvalue=[], typelist=["str"], sequence=1,
                            doc="""
Other requirements, given as a list of strings, for example:
[ 'OSTYPE == "SLC4"', '(POOL == "GENERAL" || POOL == "GEN_FARM")' ];
the final requirement is the AND of all elements in the list
""" )
    })

    _category = 'condor_requirements'
    _name = 'CondorRequirements'

    def __init__(self):
        super(CondorRequirements, self).__init__()

    def convert(self):
        '''Convert the condition(s) to a JDL specification'''

        requirementList = []

        if self.machine:
            if isinstance(self.machine, str):
                machineList = self.machine.split()
            else:
                machineList = self.machine
            machineConditionList = []
            for machine in machineList:
                machineConditionList.append("Machine == \"%s\"" % str(machine))
            machineConditionString = " || ".join(machineConditionList)
            requirement = (" ".join(["(", machineConditionString, ")"]))
            requirementList.append(requirement)

        if self.excluded_machine:
            if isinstance(self.excluded_machine, str):
                machineList = self.excluded_machine.split()
            else:
                machineList = self.excluded_machine
            for machine in machineList:
                requirementList.append("Machine != \"%s\"" % str(machine))

        if self.opsys:
            requirementList.append("OpSys == \"%s\"" % str(self.opsys))

        if self.arch:
            requirementList.append("Arch == \"%s\"" % str(self.arch))

        if self.memory:
            requirementList.append("Memory >= %s" % str(self.memory))

        if self.virtual_memory:
            requirementList.append\
                ("VirtualMemory >= %s" % str(self.virtual_memory))

        if self.other:
            requirementList.extend(self.other)

        requirementString = "requirements = " + " && ".join(requirementList)

        return requirementString

# Allow property values to be either string or list
config = getConfig("defaults_CondorRequirements")
for property in ["machine", "excluded_machine"]:
    config.options[property].type = [str, list]



from __future__ import print_function

# For running inspections of the environment
import sys
import socket
import traceback
import collections
SocketAddress = collections.namedtuple('SocketAddress', ['address', 'port'])

end_trans = '###END-TRANS###'
socket_addr = SocketAddress(address='localhost', port=5000)
server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
server_socket.bind(socket_addr)
server_socket.listen(5)
server_stdout = sys.stdout


class socketWrapper(object):

    def __init__(self, skt):
        self._socket = skt
        sys.stdout = self
        sys.stderr = self

    def write(self, stuff):
        self._socket.sendall(stuff)

    def read(self):
        cmd = ''
        while end_trans not in cmd:
            data = self._socket.recv(1024)
            if not data:
                cmd = '###BROKEN###'
                break
            cmd += data
        if cmd == '###BROKEN###':
            return ''
        return cmd.replace(end_trans, '')

# print "TCPServer Waiting for client on port 5000"


while True:
    client_socket, address = server_socket.accept()
    print("Connection request from ", SocketAddress(*address), file=server_stdout)

    sock = socketWrapper(client_socket)

    cmd = sock.read()
    if (cmd == 'q' or cmd == 'Q' or cmd == 'exit' or cmd == 'exit()'):
        client_socket.shutdown(socket.SHUT_RDWR)
        client_socket.close()
        break

    print("Executing command '%s'" % cmd, file=server_stdout)
    try:
        print(eval(cmd))
    except:
        try:
            exec(cmd)
        except:
            print("Exception raised executing command (cmd) '%s'\n" % cmd)
            print(traceback.format_exc())
    client_socket.sendall(end_trans)
    client_socket.shutdown(socket.SHUT_RDWR)
    client_socket.close()
server_socket.shutdown(socket.SHUT_RDWR)
server_socket.close()




from __future__ import absolute_import
from .Config import getConfig, makeConfig, ConfigError, configure, allConfigs, setConfigOption, expandConfigPath, config_scope, setSessionValue, getFlavour
import os.path

## from Config import getConfigDict

# here are some useful option filters


def expandvars(c, v):
    """The ~ and $VARS are automatically expanded. """
    return os.path.expanduser(os.path.expandvars(v))


def expandgangasystemvars(c, v):
    """Expands vars with the syntax '@{VAR}' from the System config item."""
    system = getConfig('System')
    for key in system.options.iterkeys():
        option = '@{%s}' % key
        if option in v:
            v = v.replace(option, system[key])
    return v



##############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: TagPrepareLocalRTHandler.py,v 1.51 2009-07-23 20:12:25 elmsheus Exp $
###############################################################################
# TagPrepare Local Runtime Handler
#
# ATLAS/ARDA

import os, pwd, commands, re, string, time, sys

from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.File import *
from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from Ganga.Utility.Config import getConfig, makeConfig, ConfigError
from Ganga.Utility.logging import getLogger

from GangaAtlas.Lib.ATLASDataset import ATLASLocalDataset

from GangaAtlas.Lib.ATLASDataset import ATLASDataset, isDQ2SRMSite, getLocationsCE, getIncompleteLocationsCE, getIncompleteLocations, whichCloud
from GangaAtlas.Lib.ATLASDataset import DQ2Dataset
from GangaAtlas.Lib.ATLASDataset import DQ2OutputDataset
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler

__directory__ = os.path.dirname(__file__)
__athdirectory__ = sys.modules['GangaAtlas.Lib.Athena'].__path__[0]

def _append_file_buffer(inputbox,name,array):

    inputbox.append(FileBuffer(name,'\n'.join(array)+'\n'))

def _append_files(inputbox,*names):

    for name in names:
        inputbox.append(File(os.path.join(__directory__,name)))

class TagPrepareLocalRTHandler(IRuntimeHandler):
    """TagPrepare Local Runtime Handler"""

    def prepare(self,app,appsubconfig,appmasterconfig,jobmasterconfig):
        """prepare the subjob specific configuration"""

        job = app._getParent() # Returns job or subjob object
        logger.debug("TagPrepareLocalRTHandler prepare called, %s", job.id)

        # prepare inputdata
        input_files = []
        input_guids = []
       
        if job.inputdata:

            # check for subjobs
            if job._getRoot().subjobs:
                if job.inputdata._name == 'ATLASLocalDataset':
                    if not job.inputdata.names: raise ApplicationConfigurationError(None,'No inputdata has been specified.')
                    input_files = job.inputdata.names

                elif job.inputdata._name == 'DQ2Dataset':
                    raise ApplicationConfigurationError(None,'Cannot use DQ2Dataset with a local job'  )
            else:
                if job.inputdata._name == 'ATLASLocalDataset':
                    input_files = ATLASLocalDataset.get_filenames(app)

                elif job.inputdata._name == 'DQ2Dataset':
                    raise ApplicationConfigurationError(None,'Cannot use DQ2Dataset with a local job'  )

        if job.outputdata:
            raise ApplicationConfigurationError(None,'No outputdata required for TagPrepare job.')

        if job._getRoot().subjobs:
            jid = "%d.%d" % (job._getRoot().id, job.id)
        else:
            jid = "%d" % job.id

        # prepare inputsandbox
        inputbox = [File(os.path.join(__athdirectory__,'athena-utility.sh')) ]
        if input_files:     _append_file_buffer(inputbox,'input_files',input_files)

        exe = os.path.join(__directory__,'run-tagprepare-local.sh')
        outputbox = jobmasterconfig.outputbox
        environment  = jobmasterconfig.env.copy()
        
        # If ArgSplitter is used
        try:
            if job.application.args:
                environment['ATHENA_OPTIONS'] = environment['ATHENA_OPTIONS'] + ' ' + ' '.join(job.application.args)
                if job.application.options:
                    job.application.options = job.application.options + ' ' + job.application.args
                else:
                    job.application.options = job.application.args
        except AttributeError:
            pass
        
        output_location = ''
        environment['OUTPUT_LOCATION'] = output_location
        environment['ATLASOutputDatasetLFC'] = config['ATLASOutputDatasetLFC']

        # Fix DATASETNAME env variable for DQ2_COPY mode
        if job.inputdata and job.inputdata._name == 'DQ2Dataset':
            if job.inputdata.dataset:
                from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import resolve_container
                datasets = resolve_container(job.inputdata.dataset)
                environment['DATASETNAME'] = datasets[0]
                try:
                    environment['DATASETLOCATION'] = ':'.join(job.inputdata.get_locations(overlap=False)[ datasets[0] ])
                except:
                    printout = 'Job submission failed ! Dataset %s could not be found in DQ2 ! Maybe retry ?' %(datasets[0])
                    raise ApplicationConfigurationError(None,printout )

        # Work around for glite WMS spaced environement variable problem
        inputbox.append(FileBuffer('athena_options',environment['ATHENA_OPTIONS']+'\n'))

        # append a property for monitoring to the jobconfig of subjobs
        lcg_config = StandardJobConfig(File(exe), inputbox, [], outputbox, environment)
        return lcg_config

    def master_prepare( self, app, appconfig):
        """Prepare the master job"""

        job = app._getParent() # Returns job or subjob object
        logger.debug('TagPrepareLCGRTHandler master_prepare called: %s', job.id )

        self.username = gridProxy.identity(safe=True)

        # prepare input sandbox
        if app.atlas_release == '':
            logger.warning('No Athena release specified - defaulting to 15.6.9')
            app.atlas_release = '15.6.9'
            
        logger.warning("Copying grid proxy to input sandbox for transfer to WN...")
                
        inputbox = [ ( File(os.path.join(__athdirectory__,'athena-utility.sh')) ),
                     ( File(os.path.join(__directory__,'get_tag_info.py')) ),
                     ( File(os.path.join(__directory__,'get_tag_info2.py')) ),
                     ( File(os.path.join(__directory__,'template.root')) ),
                     ( File(gridProxy.location())) ]
            
        ## insert more scripts to inputsandbox for FileStager
        if job.inputdata and job.inputdata._name == 'DQ2Dataset' and job.inputdata.type in ['FILE_STAGER']:
            _append_files(inputbox,'make_filestager_joption.py','dm_util.py','fs-copy.py')
            #_append_files(inputbox,'make_filestager_joption.py','dm_util.py')

        if job.inputsandbox: inputbox += job.inputsandbox
        
        #       prepare environment

        try:
            atlas_software = config['ATLAS_SOFTWARE']
        except ConfigError:
            raise ConfigError('No default location of ATLAS_SOFTWARE specified in the configuration.')

        # prepare environment
        environment={
            'MAXNUMREFS'     : str(app.max_num_refs),
            'STREAM_REF'     : app.stream_ref,
            'ATLAS_RELEASE'  : app.atlas_release,
            'ATHENA_OPTIONS' : '',
            'ATLAS_SOFTWARE' : atlas_software,
            'ATHENA_USERSETUPFILE' : '',
            'ATLAS_PROJECT' : '',
            'ATLAS_EXETYPE' : 'ATHENA',
            'GANGA_GLITE_UI': getConfig('LCG')['GLITE_SETUP'],
            'DQ2_SETUP'     : getConfig('defaults_DQ2SandboxCache')['setup'],
            'GANGA_VERSION' : configSystem['GANGA_VERSION'],
            'PROXY_NAME'    : os.path.basename(gridProxy.location()),
            'GANGA_OUTPUT_PATH' : job.outputdir
        }

        if app.lcg_prepare:
            environment['LCG_PREPARE'] = '1'
            
#       jobscript

        exe = os.path.join(__directory__,'run-tagprepare-local.sh')

#       output sandbox
        if app.lcg_prepare:        
            outputbox = [
                'taginfo.pkl', 'subcoll.tar.gz'
                ]
        else:
            outputbox = ['taginfo.pkl' ]


        if job.outputsandbox: outputbox += job.outputsandbox

        return StandardJobConfig(File(exe),inputbox,[],outputbox,environment) 

from Ganga.GPIDev.Credentials import GridProxy
gridProxy = GridProxy()

allHandlers.add('TagPrepare','Local',TagPrepareLocalRTHandler)
allHandlers.add('TagPrepare','PBS',TagPrepareLocalRTHandler)
allHandlers.add('TagPrepare','LSF',TagPrepareLocalRTHandler)

config = getConfig('Athena')
configDQ2 = getConfig('DQ2')
configSystem = getConfig('System')
logger = getLogger()



from Ganga.GPIDev.Adapters.ISplitter import SplittingError
#from GangaDirac.BOOT                    import dirac_ganga_server
from GangaDirac.Lib.Utilities.DiracUtilities import execute
from GangaDirac.Lib.Backends.DiracUtils import result_ok
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
logger = getLogger()


def igroup(iterable, num, leftovers=False):
    '''
    Generator producing sequential groups of size num from input iterable
    '''
    size = len(iterable)
    for i in xrange(0, size, num):
        if i + num > size and not leftovers:
            return
        yield iterable[i: i + num]


def DiracSizeSplitter(inputs, filesPerJob, maxSize, ignoremissing):
    """
    Generator that yields a datasets for LHCbdirac split jobs by size
    """
    #logger.debug( "DiracSplitter" )
    #logger.debug( "inputs: %s" % str( inputs ) )
    split_files = []
    i = inputs.__class__()

    if inputs.getLFNs() != len(inputs.files):
        raise SplittingError(
            "Error trying to split dataset using DIRAC backend with non-DiracFile in the inputdata")

    all_files = igroup(inputs.files[:filesPerJob], getConfig('DIRAC')['splitFilesChunks'],
                       leftovers=True)

    #logger.debug( "Looping over all_files" )
    #logger.debug( "%s" % str( all_files ) )

    for files in all_files:

        i.files = files

        LFNsToSplit = i.getLFNs()

        if(len(LFNsToSplit)) > 1:

            result = execute('splitInputDataBySize(%s,%d)'
                             % (i.getLFNs(), filesPerJob))

            if not result_ok(result):
                logger.error('DIRAC:: Error splitting files: %s' % str(result))
                raise SplittingError('Error splitting files.')

            split_files += result.get('Value', [])

        else:

            split_files = [LFNsToSplit]

    if len(split_files) == 0:
        raise SplittingError('An unknown error occured.')

    # FIXME
    # check that all files were available on the grid
    big_list = []
    for l in split_files:
        big_list.extend(l)
    diff = set(inputs.getFileNames()[:filesPerJob]).difference(big_list)
    if len(diff) > 0:
        for f in diff:
            logger.warning('Ignored file: %s' % f)
        if not ignoremissing:
            raise SplittingError('Some files not found!')
    ###

    logger.debug("Split Files: %s" % str(split_files))

    for dataset in split_files:
        yield dataset



import datetime
import os
import random
import tempfile
import time

def generateUniqueTempFile( ext = '.txt' ):

    myFile = tempfile.NamedTemporaryFile(mode='w', delete=False)

    t = datetime.datetime.now()
    unix_t = time.mktime(t.timetuple())

    file_string = str(unix_t) + "\n"

    random.seed( unix_t )
    rand = random.random() * 1E10

    file_string = file_string + str( rand ) + "\n"

    urand = os.urandom(20)

    file_string = file_string + str( urand ) + "\n"

    myFile.write( file_string )

    myFile.close()

    returnableName = myFile.name+str(ext)

    os.rename( myFile.name, returnableName )

    return returnableName





from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import Schema, Version

# Dataset class represents the empty dataset and is a base class for specific, non-empty datasets.
#
# Derived dataset classes may be freely defined and may be either specific to applications or more generic.
# The schema and internal interface is a convention between the dataset provider and the application provider.
#
# Applications are encouraged to test and handle empty datasets with the
# isEmpty() method.


class Dataset(GangaObject):
    _schema = Schema(Version(1, 0), {})
    _category = 'datasets'
    _name = "EmptyDataset"

    def __init__(self):
        super(Dataset, self).__init__()

    # Return true if the dataset is an instance of the default base class.
    # You may override it in your dataset definition but it is not mandatory.
    def isEmpty(self):
        return self._name == Dataset._name

#
#
#
# $Log: not supported by cvs2svn $
# Revision 1.2  2005/08/10 14:35:57  andrew
# Bugfix, wrong object used in return statement
#
#
#



from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Tasks import Task
from AnaTransform import AnaTransform
from Ganga.GPIDev.Schema import *
from Ganga.Utility.ColourText import ANSIMarkup, fgcol, Effects
markup = ANSIMarkup()
fx = Effects()


from Ganga.Core.exceptions import ApplicationConfigurationError

from GangaAtlas.Lib.Credentials.ProxyHelper import getNickname 


o = [""]
def c(s):
   return markup(s,fgcol("blue"))
o.append(markup(" *** Task for Athena Analysis ***", fgcol("blue")))
o.append(" Analysis Application     : "+c("t.analysis.application"))
o.append(" Set Dataset              : "+c("t.setDataset('my.dataset')"))
o.append(" Input Dataset Object     : "+c("t.analysis.inputdata"))
o.append(" Output Dataset Object    : "+c("t.analysis.outputdata"))
o.append(" Files processed per job  : "+c("t.analysis.files_per_job = 10"))
o.append("")
o.append(markup("Procedure to do a usual analysis:", fgcol("red")))
#o.append("config.Tasks.merged_files_per_job = 1 # default files per job for merged datasets")
#o.append("config.Tasks.recon_files_per_job = 10 # default files per job for recon (non-merged) datasets")
o.append("t = AnaTask()")
o.append('t.name = "MyAnalysisR1"')
o.append("t.analysis.outputdata.outputdata  = ['nTuple.root' ]")
o.append("t.analysis.application.option_file = ['./myTopOptions.py' ]")
o.append("t.analysis.application.prepare()")
o.append("t.float = 10")
o.append('t.initializeFromDatasets(["user08.MyName.ganga.dataset.recon.AOD"])')
o.append("#t.info() # Check here if settings are correct")
o.append("t.run()")
o.append("t.overview() # Watch the processing")
o.append("")
o.append("A container dataset with all outputs will be created as")
o.append(c("user.YourNickname.<YYYYMMDDHHMMSS>.task_<id>.<task name>/"))
o.append("Subtask (Transform) outputs will be put into these datasets:")
o.append(c("user.YourNickname.<YYYYMMDDHHMMSS>.task_<id>.subtask_<nr>.<inputdataset name>"))
task_help = "\n".join(o)
task_help_nocolor = task_help.replace(fgcol("blue"),"").replace(fx.normal, "").replace(fgcol("red"),"")

from dq2.clientapi.DQ2 import DQ2, DQUnknownDatasetException
dq2=DQ2(force_backend='rucio')


class AnaTask(Task):
   __doc__ = task_help_nocolor
   _schema = Schema(Version(1,1), dict(Task._schema.datadict.items() + {
        'analysis': SimpleItem(defvalue=None, transient=1, typelist=["object"], doc='Analysis Transform'),
        'container_name': SimpleItem(defvalue="",protected=True,transient=1, getter="get_container_name", doc='name of the output container'),
       }.items()))
   _category = 'tasks'
   _name = 'AnaTask'
   _exportmethods = Task._exportmethods + ["initializeFromDatasets"]

   def initialize(self):
      super(AnaTask, self).initialize()
      analysis = AnaTransform()
      analysis.exclude_from_user_area=["*.root*"]
      analysis.name = "Analysis"
      self.transforms = [analysis]
      self.setBackend(None)

   def get_container_name(self):
      name_base = ["user",getNickname(),self.creation_date,"task_%s" % self.id]
      return ".".join(name_base + [self.name]) + "/"

   def initializeFromDatasets(self,dataset_list):
      """ For each dataset in the dataset_list a transform is created. 
          The output dataset names are set using the run numbers and tags of the input datasets appended to the current t.analysis.outputdata.datasetname field."""
      if not type(dataset_list) is list:
         logger.error("dataset_list must be a python list: ['ds1','ds2',...]")
         return
      if self.analysis:
         if self.analysis.application.user_area.name:
            trf = stripProxy(self.analysis)
         else:
            logger.error("Cannot find user_area in Athena application of first transform! Have you run application.prepare()?")
            return
      else:
         logger.error("Could not find 'template' analysis transform! Create a new AnaTask and configure t.analysis, then try again.")
         return
      tid_list = []
      for dset in dataset_list:
         dset = dset.strip()
         try:
            if "*" in dset:
               logger.error("WARNING: Wildcards may include unexpected datasets in your processing! Please list your datasets before specifying them here!")
            try:
               if dset[-1] == "/":
                  tid_datasets = dq2.listDatasetsInContainer(dset)
               else:
                  tid_datasets = dq2.listDatasetsInContainer(dset+"/")
            except DQUnknownDatasetException:
               dslist = dq2.listDatasets(dset).keys()
               if len(dslist) == 0:
                  logger.error("Dataset %s not found!" % dset)
                  return
               tid_datasets = [ds for ds in dslist if "_tid" in ds and not "_sub" in ds]
               if len(tid_datasets) == 0:
                  if len(dslist) > 1:
                     logger.error("Found no tid dataset but multiple datasets match %s*!" % dset)
                     return
                  tid_datasets = [dslist[0]]
         except Exception as e:
            logger.error('DQ2 Error while listing dataset %s*! %s' % (dset, e))
            return
         logger.info("Found %i datasets matching %s..." % (len(tid_datasets), dset))
         tid_list.extend(tid_datasets)

      new_tfs = []

      for tid in tid_list:
         # check if dataset is empty
         if dq2.getNumberOfFiles(tid) == 0:
            logger.warning("Dataset %s is empty - no subtask created!" % tid)
            continue

         tf = trf.clone()
         tf.name = ".".join(tid.split(".")[1:3])

         #if "merge" in tid:
         #   tf.files_per_job = config["merged_files_per_job"]
         #   logger.warning("Files per job for %s set to %i - use 'config.Tasks.merged_files_per_job = 42' to change this value!" % (tid,tf.files_per_job))
         #else:
         #   tf.files_per_job = config["recon_files_per_job"]
         #   logger.warning("Files per job for %s set to %i - use 'config.Tasks.recon_files_per_job = 42' to change this value!" % (tid,tf.files_per_job))
         tf.inputdata.dataset=tid
         new_tfs.append(tf)
      self.transforms = new_tfs
      self.initAliases()

   def startup(self):
      super(AnaTask,self).startup()
      self.initAliases()

   def check(self):
      self.initAliases()
      if not self.name.replace(".","").replace("_","").isalnum(): # accept . and _
         logger.error("Invalid character in task name! Task names are now used for DQ2 datasets; so no spaces, slashes or other special characters are allowed.")
         raise ApplicationConfigurationError(None, "Invalid Task name!")
      super(AnaTask,self).check()

   def initAliases(self):
      self.analysis = None
      if len(self.transforms) == 1:
         self.analysis = self.transforms[0]

   def help(self):
      print task_help

   def overview(self):
      super(AnaTask, self).overview()
      print
      print "container of transform output datasets: %s" % self.container_name



from __future__ import absolute_import
##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: MassStorageFile.py,v 0.1 2011-11-09 15:40:00 idzhunov Exp $
##########################################################################
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem

from Ganga.Utility.Config import getConfig
from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory
from Ganga.GPIDev.Base.Proxy import stripProxy
from Ganga.Utility import Shell
from Ganga.Utility.logging import getLogger
from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
from Ganga.GPIDev.Base.Proxy import getName

from Ganga.GPIDev.Lib.File import FileUtils

import errno
import re
import os
import copy
import Ganga.Utility.Config

regex = re.compile('[*?\[\]]')
logger = getLogger()

class MassStorageFile(IGangaFile):

    """MassStorageFile represents a class marking a file to be written into mass storage (like Castor at CERN)
    """
    _schema = Schema(Version(1, 1), {'namePattern': SimpleItem(defvalue="", doc='pattern of the file name'),
                                     'localDir': SimpleItem(defvalue="", copyable=1, doc='local dir where the file is stored, used from get and put methods'),
                                     'joboutputdir': SimpleItem(defvalue="", doc='outputdir of the job with which the outputsandbox file object is associated'),
                                     'locations': SimpleItem(defvalue=[], copyable=1, typelist=['str'], sequence=1, doc="list of locations where the outputfiles are uploaded"),
                                     'outputfilenameformat': SimpleItem(defvalue=None, typelist=['str', 'type(None)'], protected=0,\
                                                    doc="keyword path to where the output should be uploaded, i.e. /some/path/here/{jid}/{sjid}/{fname},\
                                                        if this field is not set, the output will go in {jid}/{sjid}/{fname} or in {jid}/{fname}\
                                                        depending on whether the job is split or not"),
                                     'inputremotedirectory': SimpleItem(defvalue=None, typelist=['str', 'type(None)'], protected=0, doc="Directory on mass storage where the file is stored"),
                                     'subfiles': ComponentItem(category='gangafiles', defvalue=[], hidden=1, typelist=['Ganga.GPIDev.Lib.File.MassStorageFile'], sequence=1, copyable=0,\
                                                    doc="collected files from the wildcard namePattern"),
                                     'failureReason': SimpleItem(defvalue="", protected=1, copyable=0, doc='reason for the upload failure'),
                                     'compressed': SimpleItem(defvalue=False, typelist=['bool'], protected=0, doc='wheather the output file should be compressed before sending somewhere')
                                     })

    _category = 'gangafiles'
    _name = "MassStorageFile"
    _exportmethods = ["location", "get", "put", "setLocation", "remove", "accessURL"]

    def __init__(self, namePattern='', localDir='', **kwds):
        """ namePattern is the pattern of the output file that has to be written into mass storage
        """
        super(MassStorageFile, self).__init__()
        self._setNamePath(_namePattern=namePattern, _localDir=localDir)
        self.locations = []

        self.shell = Shell.Shell()

    def __construct__(self, args):
        if len(args) == 1 and isinstance(args[0], str):
            self._setNamePath(args[0], '')
        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], str):
            self._setNamePath(args[0], args[1])
        self.locations = []
        self.shell = Shell.Shell()

    def _setNamePath(self, _namePattern='', _localDir=''):
        if _namePattern != '' and _localDir == '':
            import os.path
            self.namePattern = os.path.basename(_namePattern)
            self.localDir = os.path.dirname(_namePattern)
        elif _namePattern != '' and _localDir != '':
            self.namePattern = _namePattern
            self.localDir = _localDir

    def _on_attribute__set__(self, obj_type, attrib_name):
        r = copy.deepcopy(self)
        if getName(obj_type) == 'Job' and attrib_name == 'outputfiles':
            r.locations = []
            r.localDir = ''
            r.failureReason = ''
        return r

    def __repr__(self):
        """Get the representation of the file."""

        return "MassStorageFile(namePattern='%s')" % self.namePattern

    def setLocation(self):
        """
        Sets the location of output files that were uploaded to mass storage from the WN
        """
        job = self.getJobObject()

        postprocessLocationsPath = os.path.join(
            job.outputdir, getConfig('Output')['PostProcessLocationsFileName'])
        if not os.path.exists(postprocessLocationsPath):
            return

        def mass_line_processor(line, mass_file):
            lineParts = line.split(' ')
            pattern = lineParts[1]
            outputPath = lineParts[2]
            name = os.path.basename(outputPath).strip('.gz')
            if regex.search(mass_file.namePattern) is not None:
                if outputPath == 'ERROR':
                    logger.error("Failed to upload file to mass storage")
                    logger.error(line[line.find('ERROR') + 5:])
                    d = MassStorageFile(namePattern=pattern)
                    d.compressed = mass_file.compressed
                    d.failureReason = line[line.find('ERROR') + 5:]
                    mass_file.subfiles.append(GPIProxyObjectFactory(d))
                else:
                    d = MassStorageFile(namePattern=name)
                    d.compressed = mass_file.compressed
                    d.outputfilenameformat = mass_file.outputfilenameformat
                    mass_file.subfiles.append(GPIProxyObjectFactory(d))
                    mass_line_processor(line, d)
            elif name == mass_file.namePattern:
                if outputPath == 'ERROR':
                    logger.error("Failed to upload file to mass storage")
                    logger.error(line[line.find('ERROR') + 5:])
                    mass_file.failureReason = line[line.find('ERROR') + 5:]
                    return
                mass_file.locations = [outputPath.strip('\n')]

        for line in open(postprocessLocationsPath, 'r'):

            if line.strip() == '':
                continue

            if line.startswith('massstorage'):
                mass_line_processor(line.strip(), self)

    def location(self):
        """
        Return list with the locations of the post processed files (if they were configured to upload the output somewhere)
        """
        tmpLocations = []
        if self.locations == []:
            if self.subfiles != []:
                for i in self.subfiles:
                    for j in i:
                        tmpLocations.append(j)
        else:
            tmpLocations = self.locations
        return tmpLocations

    def get(self):
        """
        Retrieves locally all files matching this MassStorageFile object pattern
        """

        to_location = self.localDir

        if not os.path.isdir(self.localDir):
            if self._getParent() is not None:
                to_location = self.getJobObject().outputdir
            else:
                logger.error(
                    "%s is not a valid directory.... Please set the localDir attribute" % self.localDir)
                return

        cp_cmd = getConfig('Output')['MassStorageFile'][
            'uploadOptions']['cp_cmd']

        for location in self.locations:
            targetLocation = os.path.join(
                to_location, os.path.basename(location))
            os.system('%s %s %s' % (cp_cmd, location, targetLocation))

    def getWNScriptDownloadCommand(self, indent):
        ## FIXME fix me for the situation of multiple files?

        script = """\n

###INDENT###os.system(\'###CP_COMMAND###\')

"""
        cp_cmd = '%s %s .' % (getConfig('Output')['MassStorageFile']['uploadOptions']['cp_cmd'], self.locations[0])

        replace_dict = { '###INDENT###' : indent, '###CP_COMMAND###' : cp_cmd }

        for k, v in replace_dict.iteritems():
            script = script.replace(str(k), str(v))

        return script

    def put(self):
        """
        Creates and executes commands for file upload to mass storage (Castor), this method will
        be called on the client
        """
        import glob
        import re

        sourceDir = ''

        # if used as a stand alone object
        if self._getParent() is None:
            if self.localDir == '':
                import os
                _CWD = os.getcwd()
                if os.path.isfile(os.path.join(_CWD, self.namePattern)):
                    sourceDir = _CWD
                else:
                    logger.warning('localDir attribute is empty, don\'t know from which dir to take the file')
                    return
            else:
                sourceDir = self.localDir

                (result, message) = self.validate()

                if result == False:
                    logger.warning(message)
                    return

        else:
            job = self.getJobObject()
            sourceDir = job.outputdir

            # if there are subjobs, the put method will be called on every subjob
            # and will upload the resulted output file
            if len(job.subjobs) > 0:
                return

        massStorageConfig = getConfig('Output')['MassStorageFile']['uploadOptions']

        mkdir_cmd = massStorageConfig['mkdir_cmd']
        cp_cmd = massStorageConfig['cp_cmd']
        ls_cmd = massStorageConfig['ls_cmd']
        massStoragePath = massStorageConfig['path']

        # create the last directory (if not exist) from the config path
        import os.path
        pathToDirName = os.path.dirname(massStoragePath)
        dirName = os.path.basename(massStoragePath)

        (exitcode, mystdout, mystderr) = self.execSyscmdSubprocess('%s %s' % (ls_cmd, pathToDirName))
        if exitcode != 0:
            self.handleUploadFailure(mystderr)
            return

        directoryExists = False
        for directory in mystdout.split('\n'):
            if directory.strip() == dirName:
                directoryExists = True
                break

        if not directoryExists:
            (exitcode, mystdout, mystderr) = self.execSyscmdSubprocess('%s %s' % (mkdir_cmd, massStoragePath))
            if exitcode != 0:
                self.handleUploadFailure(mystderr)
                return

        # the folder part of self.outputfilenameformat
        folderStructure = ''
        # the file name part of self.outputfilenameformat
        filenameStructure = ''

        if self._getParent() != None:
            jobfqid = self.getJobObject().fqid

            jobid = jobfqid
            subjobid = ''

            if (jobfqid.find('.') > -1):
                jobid = jobfqid.split('.')[0]
                subjobid = jobfqid.split('.')[1]

            if self.outputfilenameformat is None:
                filenameStructure = '{fname}'
                # create jid/sjid directories
                folderStructure = jobid
                if subjobid != '':
                    folderStructure = os.path.join(jobid, subjobid)

            else:
                filenameStructure = os.path.basename(self.outputfilenameformat)
                filenameStructure = filenameStructure.replace('{jid}', jobid)

                folderStructure = os.path.dirname(self.outputfilenameformat)
                folderStructure = folderStructure.replace('{jid}', jobid)

                if subjobid != '':
                    filenameStructure = filenameStructure.replace('{sjid}', subjobid)
                    folderStructure = folderStructure.replace('{sjid}', subjobid)
        else:
            if self.outputfilenameformat != None:
                folderStructure = os.path.dirname(self.outputfilenameformat)
                filenameStructure = os.path.basename(self.outputfilenameformat)
            else:
                filenameStructure = '{fname}'

        # create the folder structure
        if folderStructure != '':

            folderStructure = folderStructure.strip('/')
            massStoragePath = os.path.join(massStoragePath, folderStructure)
            command = '%s -p %s' % (mkdir_cmd, massStoragePath)
            (exitcode, mystdout, mystderr) = self.execSyscmdSubprocess(command)
            if exitcode != 0:
                self.handleUploadFailure(mystderr)
                return

        # here filenameStructure has replaced jid and sjid if any, and only not
        # replaced keyword is fname
        fileName = self.namePattern
        if self.compressed:
            fileName = '%s.gz' % self.namePattern

        if regex.search(fileName) is not None:
            for currentFile in glob.glob(os.path.join(sourceDir, fileName)):
                finalFilename = filenameStructure.replace('{fname}', os.path.basename(currentFile))
                (exitcode, mystdout, mystderr) = self.execSyscmdSubprocess('%s %s %s' % (cp_cmd, currentFile, os.path.join(massStoragePath, finalFilename)))

                d = MassStorageFile(namePattern=os.path.basename(currentFile))
                d.compressed = self.compressed

                if exitcode != 0:
                    self.handleUploadFailure(mystderr)
                else:
                    logger.info('%s successfully uploaded to mass storage as %s' % (currentFile, os.path.join(massStoragePath, finalFilename)))
                    d.locations = os.path.join(massStoragePath, os.path.basename(finalFilename))

                    # Alex removed this as more general approach in job.py after put() is called
                    # remove file from output dir if this object is attached to a job
                    # if self._getParent() != None:
                    #    os.system('rm %s' % os.path.join(sourceDir, currentFile))

                self.subfiles.append(GPIProxyObjectFactory(d))
        else:
            currentFile = os.path.join(sourceDir, fileName)
            finalFilename = filenameStructure.replace('{fname}', os.path.basename(currentFile))
            (exitcode, mystdout, mystderr) = self.execSyscmdSubprocess('%s %s %s' % (cp_cmd, currentFile, os.path.join(massStoragePath, finalFilename)))
            if exitcode != 0:
                self.handleUploadFailure(mystderr)
            else:
                logger.info('%s successfully uploaded to mass storage as %s' % (currentFile, os.path.join(massStoragePath, finalFilename)))
                location = os.path.join(massStoragePath, os.path.basename(finalFilename))
                if location not in self.locations:
                    self.locations.append(location)

                # Alex removed this as more general approach in job.py after put() is called
                # remove file from output dir if this object is attached to a job
                # if self._getParent() != None:
                #    os.system('rm %s' % os.path.join(sourceDir, currentFile))

    def validate(self):

        # if the user has set outputfilenameformat, validate for presence of
        # jid, sjid and fname keywords depending on job type - split or
        # non-split
        if self.outputfilenameformat != None:

            searchFor = ['{fname}']
            isJob = False
            isSplitJob = False

            if self._getParent() != None:

                isJob = True

                if stripProxy(self.getJobObject()).master is not None:

                    isSplitJob = True
                    searchFor.append('{sjid}')

            missingKeywords = []

            for item in searchFor:
                if self.outputfilenameformat.find(item) == -1:
                    missingKeywords.append(item)

            if len(missingKeywords):
                return (False, 'Error in MassStorageFile.outputfilenameformat field : missing keywords %s ' % ','.join(missingKeywords))

            if isSplitJob == False and self.outputfilenameformat.find('{sjid}') > -1:
                return (False, 'Error in MassStorageFile.outputfilenameformat field :  job is non-split, but {\'sjid\'} keyword found')

            if isJob == False and self.outputfilenameformat.find('{sjid}') > -1:
                return (False, 'Error in MassStorageFile.outputfilenameformat field :  no parent job, but {\'sjid\'} keyword found')

            if isJob == False and self.outputfilenameformat.find('{jid}') > -1:
                return (False, 'Error in MassStorageFile.outputfilenameformat field :  no parent job, but {\'jid\'} keyword found')

            invalidUnixChars = ['"', ' ']
            test = self.outputfilenameformat.replace('{jid}', 'a').replace('{sjid}', 'b').replace('{fname}', 'c')

            for invalidUnixChar in invalidUnixChars:
                if test.find(invalidUnixChar) > -1:
                    return (False, 'Error in MassStorageFile.outputfilenameformat field :  invalid char %s found' % invalidUnixChar)

        return (True, '')

    def handleUploadFailure(self, error):

        self.failureReason = error
        if self._getParent() != None:
            logger.error("Job %s failed. One of the job.outputfiles couldn't be uploaded because of %s" % (
                str(self._getParent().fqid), self.failureReason))
        else:
            logger.error("The file can't be uploaded because of %s" % (self.failureReason))

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """
        massStorageCommands = []

        massStorageConfig = getConfig('Output')['MassStorageFile']['uploadOptions']

        for outputFile in outputFiles:

            outputfilenameformat = 'None'
            if outputFile.outputfilenameformat != None and outputFile.outputfilenameformat != '':
                outputfilenameformat = outputFile.outputfilenameformat

            massStorageCommands.append(['massstorage', outputFile.namePattern, outputfilenameformat,
                                        massStorageConfig['mkdir_cmd'],  massStorageConfig['cp_cmd'],
                                        massStorageConfig['ls_cmd'], massStorageConfig['path']])

        import inspect
        script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                        'scripts/MassStorageFileWNScript.py')

        from Ganga.GPIDev.Lib.File import FileUtils
        script = FileUtils.loadScript(script_location, '###INDENT###')

        jobfqid = self.getJobObject().fqid

        jobid = jobfqid
        subjobid = ''

        if (jobfqid.find('.') > -1):
            jobid = jobfqid.split('.')[0]
            subjobid = jobfqid.split('.')[1]

        replace_dict = {'###MASSSTORAGECOMMANDS###' : repr(massStorageCommands),
                        '###PATTERNSTOZIP###' : str(patternsToZip),
                        '###INDENT###' : indent,
                        '###POSTPROCESSLOCATIONSFP###' : postProcessLocationsFP,
                        '###FULLJOBDIR###' : str(jobfqid.replace('.', os.path.sep)),
                        '###JOBDIR###' : str(jobid),
                        '###SUBJOBDIR###' : str(subjobid)}

        for k, v in replace_dict.iteritems():
            script = script.replace(str(k), str(v))

        return script

    def processWildcardMatches(self):
        if self.subfiles:
            return self.subfiles

        from fnmatch import fnmatch

        if regex.search(self.namePattern):
            ls_cmd = getConfig('Output')['MassStorageFile']['uploadOptions']['ls_cmd']
            exitcode, output, m = self.shell.cmd1(ls_cmd + ' ' + self.inputremotedirectory, capture_stderr=True)

            for filename in output.split('\n'):
                if fnmatch(filename, self.namePattern):
                    subfile = MassStorageFile(namePattern=filename)
                    subfile.inputremotedirectory = self.inputremotedirectory

                    self.subfiles.append(GPIProxyObjectFactory(subfile))

    def remove(self, force=False, removeLocal=False):
        """
        Removes file from remote storage ONLY by default
        """
        massStorageConfig = getConfig('Output')['MassStorageFile']['uploadOptions']
        rm_cmd = massStorageConfig['rm_cmd']

        if force == True:
            _auto_delete = True
        else:
            _auto_delete = False

        for i in self.locations:

            if not _auto_delete:

                keyin = None

                while keyin is None:
                    keyin = raw_input("Do you want to delete file %s at Location: %s ? [y/n] " % (str(self.namePattern), str(i)))
                    if keyin == 'y':
                        _delete_this = True
                    elif keyin == 'n':
                        _delete_this = False
                    else:
                        logger.warning("y/n please!")
                        keyin = None
            else:
                _delete_this = True

            if _delete_this:
                logger.info("Deleting File at Location: %s")
                self.execSyscmdSubprocess('%s %s' % (rm_cmd, i))
                self.locations.pop(i)

        if removeLocal:

            sourceDir = ''
            if self.localDir == '':
                import os
                _CWD = os.getcwd()
                if os.path.isfile(os.path.join(_CWD, self.namePattern)):
                    sourceDir = _CWD
            else:
                sourceDir = self.localDir

            _localFile = os.path.join(sourceDir, self.namePattern)

            if os.path.isfile(_localFile):

                if force:
                    _actual_delete = True
                else:

                    keyin = None
                    while keyin is None:
                        keyin = raw_input(
                            "Do you want to remove the local File: %s ? ([y]/n) " % str(_localFile))
                        if keyin in ['y', '']:
                            _actual_delete = True
                        elif keyin == 'n':
                            _actual_delete = False
                        else:
                            logger.warning("y/n please!")
                            keyin = None

                if _actual_delete:
                    import time
                    remove_filename = _localFile + "_" + str(time.time()) + '__to_be_deleted_'

                    try:
                        os.rename(_localFile, remove_filename)
                    except OSError as err:
                        logger.warning("Error in first stage of removing file: %s" % remove_filename)
                        remove_filename = _localFile

                    try:
                        os.remove(remove_filename)
                    except OSError as err:
                        if err.errno != errno.ENOENT:
                            logger.error("Error in removing file: %s" % str(remove_filename))
                            raise
                        pass
        return

    def accessURL(self):

        # Need to come up with a prescription based upon the server address and
        # file on EOS or elsewhere to return a full URL which we can pass to
        # ROOT...

        protoPath = getConfig('Output')['MassStorageFile']['defaultProtocol']

        myLocations = self.location()

        accessURLs = []

        for file in myLocations:
            import os
            accessURLs.append(protoPath + os.path.join(os.sep, file))

        return accessURLs

# add MassStorageFile objects to the configuration scope (i.e. it will be
# possible to write instatiate MassStorageFile() objects via config file)
Ganga.Utility.Config.config_scope['MassStorageFile'] = MassStorageFile




import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()

logger.debug("Loading IApplication")
import Ganga.GPIDev.Adapters.IApplication
logger.debug("Loading IBackend")
import Ganga.GPIDev.Adapters.IBackend
logger.debug("Loading ISplitter")
import Ganga.GPIDev.Adapters.ISplitter
logger.debug("Loading IMerger")
import Ganga.GPIDev.Adapters.IMerger

logger.debug("Loading GangaList")
import Ganga.GPIDev.Lib.GangaList
logger.debug("Loading File")
import Ganga.GPIDev.Lib.File
logger.debug("Loading Job")
import Ganga.GPIDev.Lib.Job

logger.debug("Loading Mergers")
import Ganga.Lib.Mergers
logger.debug("Loading Splitters")
import Ganga.Lib.Splitters

logger.debug("Loading Executable")
import Ganga.Lib.Executable
logger.debug("Loading Root")
import Ganga.Lib.Root

logger.debug("Loading LocalHost")
import Ganga.Lib.Localhost
logger.debug("Loading LCG")
import Ganga.Lib.LCG
logger.debug("Loading Condor")
import Ganga.Lib.Condor
logger.debug("Loading Interactive")
import Ganga.Lib.Interactive
logger.debug("Loading Batch")
import Ganga.Lib.Batch
logger.debug("Loading Remote")
import Ganga.Lib.Remote

logger.debug("Loading Tasks")
import Ganga.GPIDev.Lib.Tasks


logger.debug("Loading Checkers")
import Ganga.Lib.Checkers
logger.debug("Loading Notifier")
import Ganga.Lib.Notifier

logger.debug("Finished Runtime.plugins")



from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Adapters.IPrepareApp import IPrepareApp

from Ganga.Utility.logging import getLogger

logger = getLogger()

from Ganga.GPIDev.Lib.File import File

class TestApplication(IPrepareApp):
    _schema = Schema(Version(1,0), {'exe':SimpleItem(defvalue='/usr/bin/env'),
                                    'derived_value' : SimpleItem(defvalue='',typelist=['str']),
                                    'sequence' : SimpleItem([],sequence=1,typelist=['str']),
                                    'file_sequence' : FileItem(defvalue=[],sequence=1),
                                    'optsfile' : FileItem(),
                                    'modified' : SimpleItem(defvalue=0),
                                    'nodefault_file_item' : FileItem(defvalue=File('a')),
                                    'args' :  SimpleItem(defvalue=["Hello World"],typelist=['str','Ganga.GPIDev.Lib.File.File'],sequence=1,doc="List of arguments for the executable. Arguments may be strings or File objects."),# introduced for RTHandler compatibility
                                    'env' : SimpleItem(defvalue={},doc='Environment'),# introduced for RTHandler compatibility
                                    'fail': SimpleItem(defvalue='',doc='Define the artificial runtime failures: "config", "prepare"'),
                                    'postprocess_mark_as_failed': SimpleItem(defvalue=False,doc='Update teh status of the job as failed in the postprocess step'),
                                    'is_prepared' : SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)','str','bool'],protected=0,comparable=1,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
                                    'hash': SimpleItem(defvalue=None, typelist=['type(None)', 'str'], hidden=1)
                                    } )
    _name = 'TestApplication'

    _exportmethods = ['modify', 'prepare']
    
    def _auto__init__(self):
        self.derived_value = 'This is an example of the derived property: ' + self.exe

        # this is to test manually (visually) the logging level mechanism
        #print("raw print: Hello from TestApplication")
        #logger.debug('Hello from TestApplication')
        #logger.info('Hello from TestApplication')
        #logger.warning('Hello from TestApplication')
        #logger.error('Hello from TestApplication')
        #logger.critical('Hello from TestApplication')
        
    def configure(self,masterappconfig):
        if self.fail == 'config':
            x = 'triggered failure during config'
            raise Exception(x)
        appcfg = ''
        if self.fail == 'prepare':
            appcfg += 'error'
        return (None,appcfg)

    def modify(self):
        self.modified = 1
        self._setDirty(1)

    def postprocess(self):

        if self.postprocess_mark_as_failed:
            logger.info('postprocess: trying to mark the job as failed')

            from Ganga.GPIDev.Adapters.IApplication import PostprocessStatusUpdate
            raise PostprocessStatusUpdate("failed") 

    def postprocess_failed(self):
        logger.info('postprocess_failed() called')

class TestAdvancedProperties(IApplication):
    _schema = Schema(Version(1,0), {'exe':SimpleItem(defvalue='/usr/bin/env'), 'exe2' : SimpleItem(defvalue='')} )
    _name = 'TestAdvancedProperties'
    
    def _object_filter__get__(self,v):
        return self.exe

    # setting a string resets both values
    def _object_filter__set__(self,v):
        if type(v) is type(''):
            t2 = TestAdvancedProperties()
            t2.exe = v
            t2.exe2 = v
            return t2

    # if exe is set then exe2 is set to the same value
    def _attribute_filter__set__(self,n,v):
        if n == 'exe':
            self.exe2 = v
        return v

    # looks like in GPI we see everything double ;-)
    def _attribute_filter__get__(self,n,v):
        return v*2

class TestAdvancedFileProperties(IApplication):
    _schema = Schema(Version(1,0), {'file_or_files':FileItem(sequence=1,defvalue=[],strict_sequence=0)})
    _name = 'TestAdvancedFileProperties'

#    # setting files = "x" is equvalent to setting files = ['x']                
#    def _attribute_filter__set__(self,n,v):
#        if n == 'files':
#            from Ganga.GPIDev.Base.Filters import allComponentFilters
#            v = allComponentFilters['files'](v,self._schema.getItem('file_or_files'))
#            if not type(v) is type([]):
#                v = [v]
#        return v

#    # looks like in GPI we see everything double ;-)
#    def _attribute_filter__get__(self,n,v):
#        return v*2
    

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

from Ganga.Lib.Executable.Executable import RTHandler

class TestRTHandler(RTHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        if appconfig.find('error') != -1:
            x = 'triggered failure during config'
            raise Exception(x) 
        return RTHandler.prepare(self,app,appconfig,appmasterconfig,jobmasterconfig)
    
allHandlers.add('TestApplication','TestSubmitter',TestRTHandler)

## FIXME: Test for advanced properties...
## j = Job()
## a = TestAdvancedProperties()
## j.application = a






_global_queues = None

def startUpQueues():
    from Ganga.Utility.logging import getLogger
    logger = getLogger()
    global _global_queues
    if _global_queues is None:
        logger.debug("Starting Queues")
        # start queues
        from Ganga.Runtime.GPIexport import exportToGPI
        from Ganga.Core.GangaThread.WorkerThreads.ThreadPoolQueueMonitor import ThreadPoolQueueMonitor
        _global_queues = ThreadPoolQueueMonitor()
        exportToGPI('queues', _global_queues, 'Objects')

        import atexit
        atexit.register((100, shutDownQueues))

    else:
        logger.error("Cannot Start queues if they've already started")

def shutDownQueues():
    from Ganga.Utility.logging import getLogger
    logger = getLogger()
    logger.debug("Shutting Down Queues system")
    global _global_queues
    try:
        if _global_queues:
            _global_queues.lock()
            _global_queues._purge_all()
            _global_queues._stop_all_threads()
    except:
        logger.warning("Error in shutting down queues thread. Likely harmless")
    _global_queues = None
    import Ganga.GPI
    delattr(Ganga.GPI, 'queues')




from Ganga import GPI
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from new import classobj
from Ganga.GPIDev.Base.Proxy import getName, stripProxy
from .common import logger

handler_map = []


def __task__init__(self):
    # This assumes TaskApplication is #1 in MRO ( the list of methods )
    baseclass = self.__class__.mro()[2]
    baseclass.__init__(self)
    # Now do a trick to convince classes to use us if they foolishly check the name
    # (this is a bug workaround)
    #self._name = baseclass._name


_app_schema = {  'id': SimpleItem(defvalue=-1, protected=1, copyable=1, splitable=1, doc='number of this application in the transform.', typelist=["int"]),
                'tasks_id': SimpleItem(defvalue="-1:-1", protected=1, copyable=1, splitable=1, doc='id of this task:transform', typelist=["str"])}.items()

_splitter_schema = { 'task_partitions': SimpleItem(defvalue=[], copyable=1, doc='task partition numbers.', typelist=["list"]),}.items()

def taskify(baseclass, name):
    smajor = baseclass._schema.version.major
    sminor = baseclass._schema.version.minor

    cat = baseclass._category

    if cat == "applications":
        schema_items = _app_schema
        taskclass = TaskApplication
    elif cat == "splitters":
        schema_items = _splitter_schema
        taskclass = TaskSplitter

    classdict = {
        "_schema": Schema(Version(smajor, sminor), dict(baseclass._schema.datadict.items() + schema_items)),
        "_category": cat,
        "_name": name,
        "__init__": __task__init__,
    }

    if '_exportmethods' in baseclass.__dict__:
        classdict['_exportmethods'] = baseclass.__dict__['_exportmethods']
    cls = classobj(name, (taskclass, baseclass), classdict)

    global handler_map
    # Use the same handlers as for the base class
    handler_map.append((getName(baseclass), name))

    return cls


class TaskApplication(object):

    def getTransform(self):
        tid = self.tasks_id.split(":")
        if len(tid) == 2 and tid[0].isdigit() and tid[1].isdigit():
            try:
                task = GPI.tasks(int(tid[0]))
            except KeyError:
                return None
            if task:
                return task.transforms[int(tid[1])]
        if len(tid) == 3 and tid[1].isdigit() and tid[2].isdigit():
            task = GPI.tasks(int(tid[1]))
            if task:
                return task.transforms[int(tid[2])]
        return None

    def transition_update(self, new_status):
        # print "Transition Update of app ", self.id, " to ",new_status
        try:
            transform = self.getTransform()
            if self.tasks_id.startswith("00"):  # Master job
                if new_status == "new":  # something went wrong with submission
                    for sj in self._getParent().subjobs:
                        sj.application.transition_update(new_status)

                if transform:
                    stripProxy(transform).setMasterJobStatus(
                        self._getParent(), new_status)

            else:
                if transform:
                    stripProxy(transform).setAppStatus(self, new_status)

        except Exception as x:
            import traceback
            import sys
            logger.error(
                "Exception in call to transform[%s].setAppStatus(%i, %s)", self.tasks_id, self.id, new_status)
            logger.error( getName(x) + " : " + x)
            tb = sys.exc_info()[2]
            if tb:
                traceback.print_tb(tb)
            else:
                logger.error("No Traceback available")

            logger.error("%s", x)


class TaskSplitter(object):
    # Splitting based on numsubjobs

    def split(self, job):
        subjobs = self.__class__.mro()[2].split(self, job)
        # Get information about the transform
        transform = stripProxy(job.application.getTransform())
        id = job.application.id
        partition = transform._app_partition[id]
        # Tell the transform this job will never be executed ...
        transform.setAppStatus(job.application, "removed")
        # .. but the subjobs will be
        for i in range(0, len(subjobs)):
            subjobs[i].application.tasks_id = job.application.tasks_id
            subjobs[i].application.id = transform.getNewAppID(
                self.task_partitions[i])
            # Do not set to submitting - failed submission will make the applications stuck...
            # transform.setAppStatus(subjobs[i].application, "submitting")
        if not job.application.tasks_id.startswith("00"):
            job.application.tasks_id = "00:%s" % job.application.tasks_id
        return subjobs


from Ganga.Lib.Executable.Executable import Executable
from Ganga.Lib.Splitters import ArgSplitter

ExecutableTask = taskify(Executable, "ExecutableTask")
ArgSplitterTask = taskify(ArgSplitter, "ArgSplitterTask")

task_map = {"Executable": ExecutableTask}


def taskApp(app):
    """ Copy the application app into a task application. Returns a task application without proxy """
    a = stripProxy(app)
    if "Task" in a._name:
        return a
    elif a._name in task_map:
        b = task_map[a._name]()

    else:
        logger.error("The application '%s' cannot be used with the tasks package yet!" % getName(a))
        raise AttributeError()
    for k in a.getNodeData():
        b.setNodeAttribute(k, a.getNodeAttribute(k))

    # We need to recalculate the application's preparable hash here, since the text string representation
    # of the application has changed (e.g. Executable -> ExecutableTask).
    if hasattr(b, 'hash') and b.hash is not None:
        try:
            b.calc_hash()
        except:
            logger.warn('Non fatal error recalculating the task application hash value')

    return b



# Tools to check the output file
# Useful!
import os
import sys
import re
import subprocess
import optparse

all_env = {}

class CheckRootException(Exception):
    def __init__(self, error):
        Exception.__init__(self, error)


class OutputChecker(object):
    '''Base class for the output checker
    '''
    def __init__(self):
        pass
    def get_root_entries(self, filename, n_entries):
        pass


class OutputCheckerDev(OutputChecker):
    '''Output checker for RAT-dev.

    Does nothing, because we have no idea which version of RAT is being used
    without checking the git log and doing loads of unreliable stuff.
    '''
    def __init__(self):
        super(OutputCheckerDev, self).__init__()


class OutputCheckerPre460(OutputChecker):
    '''Output checker for RAT-4.5.0 and earlier.
    '''
    def __init__(self):
        super(OutputCheckerPre460, self).__init__()
    def get_root_entries(self, filename, n_entries):
        import ROOT
        import rat
        try:
            tf = ROOT.TFile(filename)
            tt = ROOT.TTree()
            # check root files (and soc files?)
            if tf.Get("T"):
                tt = tf.Get("T")
            # check ntuples
            elif tf.Get("output"):
                tt = tf.Get("output")
            else:
                raise CheckRootException("Neither TTree 'T' nor TTree 'output' exist")

            # check entries in TTrees
            if n_entries == tt.GetEntries():
                pass
            else:
                raise CheckRootException("Number of events simulated is incorrect")
        except Exception as e:
            raise CheckRootException("Cannot get TTree: %s" % e)
    

class OutputCheckerPost460(OutputChecker):
    '''Output checker for RAT-4.6.0 and later.
    '''
    def __init__(self):
        super(OutputCheckerPost460, self).__init__()
    def get_root_entries(self, filename, n_entries):
        import ROOT
        import rat
        try:
            # check standard root files
            if rat.dsreader(filename):
                for ds, run in rat.dsreader(filename):
                    if n_entries == run.GetNumberOfEventsSimulated(run):
                        break # assume simulations only do one run
                    else:
                        raise CheckRootException("Number of events simulated is incorrect")

            #check soc files
            elif rat.socreader(filename):
                for soc, run in rat.socreader(filename):
                    if n_entries == run.GetNumberOfEventsSimulated(run):
                        break # assume simulations only do one run
                    else:
                        raise CheckRootException("Number of events simulated is incorrect")
            #check ntuples
            else:
                tf = ROOT.TFile(filename)
                tt = tf.Get("output")

                # check entries in TTrees
                if n_entries == tt.GetEntries():
                    pass
                else:
                    raise CheckRootException("Number of events simulated is incorrect")            
        except Exception as e:
            raise CheckRootException("Cannot get TTree: %s" % e)


def get_checker(rat_version):
    try:
        (major, minor, patch) = (int(s) for s in rat_version.split('.'))
        if major>4 or (major==4 and minor>5):
            return OutputCheckerPost460()
        else:
            return OutputCheckerPre460()
    except ValueError:
        if rat_version=='dev':
            return OutputCheckerDev()
        else:
            raise CheckRootException("Unknown rat version: %s" % rat_version)


if __name__ == '__main__':
    parser = optparse.OptionParser()
    parser.add_option("-f", dest="filename")
    parser.add_option("-n", dest="n_entries")
    parser.add_option("-v", dest="rat_version")
    (options, args) = parser.parse_args()

    output_checker = get_checker(options.rat_version)
    output_checker.get_root_entries(options.filename, int(options.n_entries))
    # Signal that checks ran fine
    open("__CHECK_ROOT_SUCCESS__", 'w').close()
